import os
import torch
import torch.optim as optim
import torch.nn as nn
from typing import Any, Dict, List, Set
import copy
import itertools
from model.semseg.deeplabv3plus import DeepLabV3Plus
from model.semseg.mask2former import Mask2Former
from model.semseg.upernet import UPerNet
from model.semseg.unetformer import UNetFormer

# from dataset import build_dataloader
from configs.args import settings
from utils.net_utils import loadWeight, seed_torch, EarlyStopping
from utils.metrics import Evaluator
from losses.criterion import SetCriterion
from losses.matcher import HungarianMatcher
from losses.celoss import SegmentationLosses

# 构建模型函数
def build_model(args: settings):
    seed_torch()
    print(f"feature encoder : {args.backbone}")
    print(f"segment decoder : {args.segHead}")
    if args.segHead == "deeplabv3":
        model = DeepLabV3Plus(backbone=args.backbone, nclass=args.nclass, isContext=args.isContext)
    elif args.segHead == "mask2former":
        model = Mask2Former(backbone=args.backbone, nclass=args.nclass, isContext=args.isContext)
    elif args.segHead == "upernet":
        model = UPerNet(backbone=args.backbone, nclass=args.nclass, isContext=args.isContext)
    elif args.segHead == "unetformer":
        model = UNetFormer(backbone=args.backbone, nclass=args.nclass, isContext=args.isContext)
    else:
        raise ValueError(f"Invalid seg head: {args.segHead}")
    if args.resume:
        loadWeight(path=args.pth_resume, net=model)
    # model.to(args.device)
    return model

# 构建数据加载器函数
def build_data(args: settings):
    print(f"data type : {args.data_type}")
    print(f"downsample : {args.downsample}")
    trainDataLoader = build_dataloader(
        data_path=os.path.join(args.data_root, f"{args.data_type}_train_d{args.downsample}_1024.h5"),
        batch_size=args.batch_size,
        num_workers=args.num_workers,
        isSingleClass=args.data_type == "FBP")
    valDataLoader = build_dataloader(
        data_path=os.path.join(args.data_root, f"{args.data_type}_val_d{args.downsample}_1024.h5"),
        batch_size=1,  # val batch size = 1
        isSingleClass=args.data_type == "FBP")

    return trainDataLoader, valDataLoader

# 构建Mask2Former优化器
def build_optimizer_mask2former(model, lr):
    weight_decay_norm = 0.0  # cfg.SOLVER.WEIGHT_DECAY_NORM
    weight_decay_embed = 0.0  # cfg.SOLVER.WEIGHT_DECAY_EMBED

    defaults = {}
    BASE_LR = lr
    defaults["lr"] = BASE_LR  # cfg.SOLVER.BASE_LR
    defaults["weight_decay"] = 0.05  # cfg.SOLVER.WEIGHT_DECAY

    norm_module_types = (
        torch.nn.BatchNorm1d,
        torch.nn.BatchNorm2d,
        torch.nn.BatchNorm3d,
        torch.nn.SyncBatchNorm,
        # NaiveSyncBatchNorm inherits from BatchNorm2d
        torch.nn.GroupNorm,
        torch.nn.InstanceNorm1d,
        torch.nn.InstanceNorm2d,
        torch.nn.InstanceNorm3d,
        torch.nn.LayerNorm,
        torch.nn.LocalResponseNorm,
    )

    params: List[Dict[str, Any]] = []
    memo: Set[torch.nn.parameter.Parameter] = set()
    for module_name, module in model.named_modules():
        for module_param_name, value in module.named_parameters(recurse=False):
            if not value.requires_grad:
                continue
            # Avoid duplicating parameters
            if value in memo:
                continue
            memo.add(value)

            hyperparams = copy.copy(defaults)
            if "backbone" in module_name:
                hyperparams["lr"] = hyperparams["lr"] * 0.1  # cfg.SOLVER.BACKBONE_MULTIPLIER
            if (
                    "relative_position_bias_table" in module_param_name
                    or "absolute_pos_embed" in module_param_name
            ):
                print(module_param_name)
                hyperparams["weight_decay"] = 0.0
            if isinstance(module, norm_module_types):
                hyperparams["weight_decay"] = weight_decay_norm
            if isinstance(module, torch.nn.Embedding):
                hyperparams["weight_decay"] = weight_decay_embed
            params.append({"params": [value], **hyperparams})

    def maybe_add_full_model_gradient_clipping(optim):
        # detectron2 doesn't have full model gradient clipping now
        clip_norm_val = 0.1  # cfg.SOLVER.CLIP_GRADIENTS.CLIP_VALUE
        enable = (
            # cfg.SOLVER.CLIP_GRADIENTS.ENABLED
            # and cfg.SOLVER.CLIP_GRADIENTS.CLIP_TYPE == "full_model"
                True
                and clip_norm_val > 0.0
        )

        class FullModelGradientClippingOptimizer(optim):
            def step(self, closure=None):
                all_params = itertools.chain(*[x["params"] for x in self.param_groups])
                torch.nn.utils.clip_grad_norm_(all_params, clip_norm_val)
                super().step(closure=closure)

        return FullModelGradientClippingOptimizer if enable else optim

    optimizer_type = "ADAMW"  # cfg.SOLVER.OPTIMIZER
    if optimizer_type == "SGD":
        optimizer = maybe_add_full_model_gradient_clipping(torch.optim.SGD)(
            params, BASE_LR, momentum=cfg.SOLVER.MOMENTUM
        )
    elif optimizer_type == "ADAMW":
        optimizer = maybe_add_full_model_gradient_clipping(torch.optim.AdamW)(
            params, BASE_LR
        )
    else:
        raise NotImplementedError(f"no optimizer type {optimizer_type}")
    # if not cfg.SOLVER.CLIP_GRADIENTS.CLIP_TYPE == "full_model":
    #     optimizer = maybe_add_gradient_clipping(cfg, optimizer)
    return optimizer

# 构建优化器和学习率调度器
def build_optimizer(net, args: settings):
    # optimizer
    # if args.segHead == "mask2former":
    #     optimizer = build_optimizer_mask2former(net, lr=args.learn_rate)
    # else:
    #     # optimizer = optim.SGD(net.parameters(), lr=args.learn_rate, momentum=0.9, weight_decay=0.0005)
    #     optimizer = optim.AdamW(net.parameters(), lr=args.learn_rate, betas=(0.9, 0.999),
    #                             eps=1e-08, weight_decay=0.05, amsgrad=False)
    optimizer = build_optimizer_mask2former(net, lr=args.learn_rate)
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'max', patience=args.lr_patience, verbose=True)
    return optimizer, scheduler

# 构建评估器和早停机制
def build_evaluator(args: settings, now_criterion: float = 0):
    evaluator = Evaluator(args.nclass, args.ignore_index)
    earlystopping = EarlyStopping(criterion=now_criterion, patience=args.early_stopping_patience)
    return evaluator, earlystopping

# 构建损失函数
def build_loss(args: settings):
    if args.segHead != "mask2former":
        if args.nclass > 2:
            weight = torch.ones(args.nclass).to(args.device)
            weight[0] = 0
        else:
            weight = None
        criterion = SegmentationLosses(weight=weight, ignore_index=args.ignore_index,
                                       device=args.device).build_loss(mode="ce")
        return criterion
    # 损失参数
    deep_supervision = args.deep_supervision
    no_object_weight = args.no_object_weight

    # 损失权重
    class_weight = args.class_weight
    dice_weight = args.dice_weight
    mask_weight = args.mask_weight

    # 构建标准
    matcher = HungarianMatcher(
        cost_class=class_weight,
        cost_mask=mask_weight,
        cost_dice=dice_weight,
        num_points=args.train_num_points,
    )

    weight_dict = {"loss_ce": class_weight, "loss_mask": mask_weight, "loss_dice": dice_weight}
    if deep_supervision:
        dec_layers = args.dec_layers
        aux_weight_dict = {}
        for i in range(dec_layers - 1):
            aux_weight_dict.update({k + f"_{i}": v for k, v in weight_dict.items()})
        weight_dict.update(aux_weight_dict)

    losses = ["labels", "masks"]
    criterion = SetCriterion(
        args.nclass,
        matcher=matcher,
        weight_dict=weight_dict,
        eos_coef=no_object_weight,
        losses=losses,
        num_points=args.train_num_points,
        oversample_ratio=args.oversample_ratio,
        importance_sample_ratio=args.importance_sample_ratio,
        device=args.device
    )
    return criterion



from model.semseg.base import BaseNet

import torch
from torch import nn
import torch.nn.functional as F
from model.semseg.GCM import Global_Context_Module


class DeepLabV3Plus(BaseNet):
    def __init__(self, backbone, nclass, isContext: bool = False):
        super(DeepLabV3Plus, self).__init__(backbone)
        self.isContext = isContext
        print(f"isContext = {self.isContext}")
        if self.isContext:
            self.GCM = Global_Context_Module(channels=self.backbone.channels)
        low_level_channels = self.backbone.channels[0]
        high_level_channels = self.backbone.channels[-1]

        self.head = ASPPModule(high_level_channels, (12, 24, 36))

        self.reduce = nn.Sequential(nn.Conv2d(low_level_channels, 48, 1, bias=False),
                                    nn.BatchNorm2d(48),
                                    nn.ReLU(True))

        self.fuse = nn.Sequential(nn.Conv2d(high_level_channels // 8 + 48, 256, 3, padding=1, bias=False),
                                  nn.BatchNorm2d(256),
                                  nn.ReLU(True),

                                  nn.Conv2d(256, 256, 3, padding=1, bias=False),
                                  nn.BatchNorm2d(256),
                                  nn.ReLU(True),
                                  nn.Dropout(0.1, False))

        self.classifier = nn.Conv2d(256, nclass, 1, bias=True)

    def base_forward0(self, x):
        h, w = x.shape[-2:]

        c1, _, _, c4 = self.backbone.base_forward(x)

        c4 = self.head(c4)
        c4 = F.interpolate(c4, size=c1.shape[-2:], mode="bilinear", align_corners=True)

        c1 = self.reduce(c1)

        out = torch.cat([c1, c4], dim=1)
        out_feat = self.fuse(out)
        # feature = F.interpolate(out_feat, size=(h, w), mode="bilinear", align_corners=True)
        out = self.classifier(out_feat)
        out = F.interpolate(out, size=(h, w), mode="bilinear", align_corners=True)
        return out

    def base_forward_context(self, xs):
        x1, x2, x3 = xs
        with torch.no_grad():
            f2 = self.backbone.base_forward(x2)
            f3 = self.backbone.base_forward(x3)
        f1 = self.backbone.base_forward(x1)
        h, w = x1.shape[-2:]

        c1, _, _, c4 = self.GCM(f1, f2, f3)

        c4 = self.head(c4)
        c4 = F.interpolate(c4, size=c1.shape[-2:], mode="bilinear", align_corners=True)

        c1 = self.reduce(c1)

        out = torch.cat([c1, c4], dim=1)
        out_feat = self.fuse(out)
        # feature = F.interpolate(out_feat, size=(h, w), mode="bilinear", align_corners=True)
        out = self.classifier(out_feat)
        out = F.interpolate(out, size=(h, w), mode="bilinear", align_corners=True)
        return out

    def base_forward(self, x):
        if self.isContext:
            return self.base_forward_context(x)
        else:
            return self.base_forward0(x)


def ASPPConv(in_channels, out_channels, atrous_rate):
    block = nn.Sequential(nn.Conv2d(in_channels, out_channels, 3, padding=atrous_rate,
                                    dilation=atrous_rate, bias=False),
                          nn.BatchNorm2d(out_channels),
                          nn.ReLU(True))
    return block


class ASPPPooling(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(ASPPPooling, self).__init__()
        self.gap = nn.Sequential(nn.AdaptiveAvgPool2d(1),
                                 nn.Conv2d(in_channels, out_channels, 1, bias=False),
                                 nn.BatchNorm2d(out_channels),
                                 nn.ReLU(True))

    def forward(self, x):
        h, w = x.shape[-2:]
        pool = self.gap(x)
        return F.interpolate(pool, (h, w), mode="bilinear", align_corners=True)


class ASPPModule(nn.Module):
    def __init__(self, in_channels, atrous_rates):
        super(ASPPModule, self).__init__()
        out_channels = in_channels // 8
        rate1, rate2, rate3 = atrous_rates

        self.b0 = nn.Sequential(nn.Conv2d(in_channels, out_channels, 1, bias=False),
                                nn.BatchNorm2d(out_channels),
                                nn.ReLU(True))
        self.b1 = ASPPConv(in_channels, out_channels, rate1)
        self.b2 = ASPPConv(in_channels, out_channels, rate2)
        self.b3 = ASPPConv(in_channels, out_channels, rate3)
        self.b4 = ASPPPooling(in_channels, out_channels)

        self.project = nn.Sequential(nn.Conv2d(5 * out_channels, out_channels, 1, bias=False),
                                     nn.BatchNorm2d(out_channels),
                                     nn.ReLU(True),
                                     nn.Dropout2d(0.5, False))

    def forward(self, x):
        feat0 = self.b0(x)
        feat1 = self.b1(x)
        feat2 = self.b2(x)
        feat3 = self.b3(x)
        feat4 = self.b4(x)
        y = torch.cat((feat0, feat1, feat2, feat3, feat4), 1)
        return self.project(y)



import torch
import torch.nn as nn
import torch.nn.functional as F
from scipy.optimize import linear_sum_assignment
from torch.cuda.amp import autocast


def calIoU(mask1, mask2, eps=10 ** -9):
    mask1 = mask1 > 0.5
    mask2 = mask2 > 0.5

    intersection = torch.logical_and(mask1.unsqueeze(1), mask2.unsqueeze(0)).sum(dim=(2, 3))
    union = torch.logical_or(mask1.unsqueeze(1), mask2.unsqueeze(0)).sum(dim=(2, 3))

    iou = intersection.float() / (union.float() + eps)

    return iou


def batch_kl_loss(inputs: torch.Tensor, targets: torch.Tensor):
    """
    Compute the KL divergence matrix between two sets of distributions.
    """
    # Apply softmax to convert logits to probabilities
    inputs = F.softmax(inputs.flatten(1), dim=-1)  # [q,c]
    targets = F.softmax(targets.flatten(1), dim=-1)  # [q,c]

    kl_matrix = torch.einsum("qxc,qmc->qm", targets[:, None, :],
                             torch.log(targets[:, None, :] + 1e-10) - torch.log(inputs + 1e-10))
    
    return kl_matrix


@torch.no_grad()
def memory_efficient_forward(outmasks1, outmasks2):
    """More memory-friendly matching"""
    bs, num_queries = outmasks1.shape[:2]

    indices = []

    # Iterate through batch size
    for b in range(bs):
        out_mask1 = outmasks1[b]
        out_mask2 = outmasks2[b]

        with autocast(enabled=False):
            # Compute the cost between masks
            cost_mask = batch_kl_loss(out_mask1, out_mask2)
            # cost_mask = calIoU(out_mask1, out_mask2)
        # print(cost_mask.shape, "\n", cost_mask)
        # Final cost matrix
        C = cost_mask
        C = C.reshape(num_queries, -1).cpu()  # [num_queries, num_total_targets]

        indices.append(linear_sum_assignment(C))

    return [
        (torch.as_tensor(i, dtype=torch.int64), torch.as_tensor(j, dtype=torch.int64))
        for i, j in indices
    ]


def crop(input, shape):
    h0, w0 = shape
    h, w = input.shape[-2:]
    dh = (h - h0) // 2
    dw = (w - w0) // 2
    return input[..., dh:dh + h0, dw:dw + h0]


class FedTrain(object):
    def __init__(self, device="cpu", deviceC="cpu", downsample_tuple=(1, 2, 4)):
        self.device = device
        self.deviceC = deviceC
        assert len(downsample_tuple) == 3
        d1, d2, d3 = downsample_tuple
        assert d1 == 1
        self.d1 = d1
        self.d2 = d2
        self.w2 = 1 / d2
        self.d3 = d3
        self.w3 = 1 / d3
        self.kl_loss = nn.KLDivLoss(reduction="batchmean")

    def CalKL(self, input, target):
        KLloss = self.kl_loss(F.log_softmax(input, dim=1), F.softmax(target, dim=1))
        return KLloss

    def Align(self, InputList):
        feat1, feat2, feat3 = InputList
        h, w = feat1.shape[-2:]
        feat2 = F.interpolate(feat2, scale_factor=self.d2, mode="bilinear", align_corners=False)
        feat3 = F.interpolate(feat2, scale_factor=self.d3, mode="bilinear", align_corners=False)
        return feat1, crop(feat2, (h, w)), crop(feat3, (h, w))

    def _get_permutation_idx1(self, indices):
        # permute idx1 following indices
        batch_idx = torch.cat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.cat([src for (src, _) in indices])
        return batch_idx, src_idx

    def _get_permutation_idx2(self, indices):
        # permute targets following indices
        batch_idx = torch.cat([torch.full_like(tgt, i) for i, (_, tgt) in enumerate(indices)])
        tgt_idx = torch.cat([tgt for (_, tgt) in indices])
        return batch_idx, tgt_idx

    def calLoss(self, indices, resA, resB):
        idxA = self._get_permutation_idx1(indices)
        # print(idxA)
        resALabel = resA["pred_logits"][idxA]
        # resAMask = resA["pred_masks"][idxA]

        idxB = self._get_permutation_idx2(indices)
        resBLabel = resB["pred_logits"][idxB]
        # resBMask = resB["pred_masks"][idxB]
        # print(resALabel.shape, resBLabel.shape, resAMask.shape, resBMask.shape)

        return self.CalKL(resALabel.to(self.device), resBLabel.to(self.device))

    def trainer(self, out1, out2, out3):
        # "pred_logits" [bs,q,nclass]
        # "pred_masks" [bs,q,h,w]

        # Align
        _, _, h, w = out1["pred_masks"].shape
        out2["pred_masks"] = crop(
            F.interpolate(out2["pred_masks"], scale_factor=self.d2, mode="bilinear", align_corners=False),
            (h, w)).to(self.deviceC)
        out3["pred_masks"] = crop(
            F.interpolate(out3["pred_masks"], scale_factor=self.d3, mode="bilinear", align_corners=False),
            (h, w)).to(self.deviceC)
        match12 = memory_efficient_forward(out1["pred_masks"].to(self.deviceC), out2["pred_masks"])
        match13 = memory_efficient_forward(out1["pred_masks"].to(self.deviceC), out3["pred_masks"])

        loss12 = self.calLoss(match12, out1, out2)
        loss13 = self.calLoss(match13, out1, out3)
        return loss12 * self.w2 + loss13 * self.w3


if __name__ == "__main__":
    a, b = torch.ones((1, 4, 4)), torch.zeros((1, 4, 4))
    a[0, 1, 1] = 0
    b[0, 0, 0] = 1
    print(a, "\n", b)
    print(calIoU(torch.cat((a, b), dim=0), torch.cat((b, a), dim=0)))
    print(memory_efficient_forward(torch.cat((a, b, a), dim=0).unsqueeze(0),
                                   torch.cat((b, a, a), dim=0).unsqueeze(0)))

    output = F.kl_div(F.log_softmax(a.flatten(1), dim=-1), F.softmax(b.flatten(1), dim=-1), reduction='batchmean')
    print(1, output)

    output = F.kl_div(F.log_softmax(b.flatten(1), dim=-1), F.softmax(a.flatten(1), dim=-1), reduction='batchmean')
    print(2, output)
    print(aa)
    s = 128
    nclass = 26
    bs = 2
    featsList, predList = [], []
    out1 = {"pred_logits": torch.randn((bs, 100, nclass), requires_grad=True),
            "pred_masks": torch.randn((bs, 100, s, s), requires_grad=True)}
    out2 = {"pred_logits": torch.randn((bs, 100, nclass)),
            "pred_masks": torch.randn(bs, 100, s, s)}
    out3 = {"pred_logits": torch.randn((bs, 100, nclass)),
            "pred_masks": torch.randn(bs, 100, s, s)}
    f = FedTrain()
    l = f.trainer(out1, out2, out3)
    print(l)


    def batch_dice_loss(inputs: torch.Tensor, targets: torch.Tensor):
        """
        Compute the DICE loss, similar to generalized IOU for masks
        Args:
            inputs: A float tensor of arbitrary shape.
                    The predictions for each example.
            targets: A float tensor with the same shape as inputs. Stores the binary
                     classification label for each element in inputs
                    (0 for the negative class and 1 for the positive class).
        """
        inputs = inputs.sigmoid()
        inputs = inputs.flatten(1)
        numerator = 2 * torch.einsum("nc,mc->nm", inputs, targets)
        denominator = inputs.sum(-1)[:, None] + targets.sum(-1)[None, :]
        loss = 1 - (numerator + 1) / (denominator + 1)
        return loss


    y = batch_dice_loss(inputs=torch.randn(100, 200), targets=torch.ones(8, 200))
    print(y.shape)

    print(torch.randn(3, 6, 8).flatten(1).shape)



from .backbone import resnet, swin_transformer, checkpoint
from .semseg.deeplabv3plus import DeepLabV3Plus


def build_model(backbone, nclass, segHead="deeplabV3P"):
    if segHead == "deeplabV3P":
        net = DeepLabV3Plus(backbone=backbone, nclass=nclass)
    else:
        raise EnvironmentError
    return net



import torch
import torch.nn as nn


class SegmentationLosses(object):
    def __init__(self, weight=None, reduction='mean', ignore_index=-100, device=torch.device("cuda")):
        self.ignore_index = ignore_index
        self.weight = weight
        self.reduction = reduction
        self.device = device

    def build_loss(self, mode='ce'):
        """Choices: ['ce' or 'focal']"""
        if mode == 'ce':
            return self.CrossEntropyLoss
        elif mode == 'focal':
            return self.FocalLoss
        else:
            raise NotImplementedError

    def CrossEntropyLoss(self, logit, target):
        criterion = nn.CrossEntropyLoss(weight=self.weight, ignore_index=self.ignore_index,
                                        reduction=self.reduction).to(self.device)

        loss = criterion(logit, target.long())
        losses = {"loss_tot": loss}
        return losses

    def FocalLoss(self, logit, target, gamma=2, alpha=0.5):
        criterion = nn.CrossEntropyLoss(weight=self.weight, ignore_index=self.ignore_index,
                                        reduction=self.reduction).to(self.device)
        logpt = -criterion(logit, target.long())
        pt = torch.exp(logpt)
        if alpha is not None:
            logpt *= alpha
        loss = -((1 - pt) ** gamma) * logpt
        losses = {"loss_tot": loss}
        return losses



from model.semseg.base import BaseNet
import torch
from torch import nn
import torch.nn.functional as F
from addict import Dict

from .utils_mask2former.pixel_decoder.msdeformattn import MSDeformAttnPixelDecoder
from .utils_mask2former.transformer_decoder.mask2former_transformer_decoder import MultiScaleMaskedTransformerDecoder


class Mask2Former(BaseNet):
    def __init__(self, backbone, nclass: int = 25):
        super(Mask2Former, self).__init__(backbone)
        self.backbone_feature_shape = dict()
        self.cal_shape()
        self.sem_seg_head = MaskFormerHead(input_shape=self.backbone_feature_shape,
                                           nclass=nclass)

        # TODO
        self.attention12 = nn.ModuleList()
        self.attention13 = nn.ModuleList()
        self.Squeue = nn.ModuleList()
        for k, v in self.backbone_feature_shape.items():
            # if k == "res2": continue
            self.attention12.append(nn.MultiheadAttention(embed_dim=v['channel'], num_heads=2))
            self.attention13.append(nn.MultiheadAttention(embed_dim=v['channel'], num_heads=2))
            self.Squeue.append(nn.Conv2d(v['channel'] * 3, v['channel'], kernel_size=1, stride=1, padding=0))

    def cal_shape(self):
        out = self.backbone.base_forward(torch.randn((2, 3, 512, 512)))
        for i, x in enumerate(out):
            self.backbone_feature_shape[f'res{i + 2}'] = Dict({'channel': x.shape[1], 'stride': 512 // x.shape[2]})
        del self.backbone_feature_shape['res2']
        print("mask2former get features shapes:")
        print(self.backbone_feature_shape)

    @staticmethod
    def att(q, kv, attention):
        # (L,N,E)
        # where L is the target sequence length, N is the batch size, E is the embedding dimension.
        # return q
        # 将图像特征展平为序列长度
        bs, c, h, w = q.shape
        q = q.flatten(2).permute(2, 0, 1)  # h*w,bs,c
        kv = kv.flatten(2).permute(2, 0, 1)

        output, _ = attention(q, kv, kv)
        output = output.permute(1, 2, 0)  # (batch_size, embed_dim, seq_len)
        output = output.view(bs, c, h, w)  # (batch_size, embed_dim, height, width)
        return output

    def base_forward(self, xs):
        x1, x2, x3 = xs

        f1 = self.backbone.base_forward(x1)[1:]
        f2 = self.backbone.base_forward(x2)[1:]
        f3 = self.backbone.base_forward(x3)[1:]

        features = dict()
        for i, x in enumerate(zip(f1, f2, f3)):
            fx1, fx2, fx3 = x
            Fx = torch.cat((fx1,
                            self.att(q=fx1, kv=fx2, attention=self.attention12[i]),
                            self.att(q=fx1, kv=fx3, attention=self.attention13[i])),
                           dim=1)
            features[f'res{i + 3}'] = self.Squeue[i](Fx)
            # print(fx1.shape, features[f'res{i + 2}'].shape)
        outputs = self.sem_seg_head(features)
        return outputs

    # 最初的正确版本
    def base_forward0(self, x):

        f = self.backbone.base_forward(x)
        features = dict()
        for i, x in enumerate(f):
            features[f'res{i + 2}'] = x
        outputs = self.sem_seg_head(features)
        return outputs

        if self.training:
            return outputs
        else:
            # delete last ==================#
            mask_cls = outputs["pred_logits"][..., 1:-1]
            mask_pred = outputs["pred_masks"]
            # mask_pred = F.interpolate(
            #     mask_pred,
            #     size=(h, w),
            #     mode="bilinear",
            #     align_corners=False,
            # )
            mask_cls = F.softmax(mask_cls, dim=-1)
            mask_pred = mask_pred.sigmoid()
            pred_mask = torch.einsum("bqc,bqhw->bchw", mask_cls, mask_pred)
            return pred_mask, outputs["mask_features"]
            # return F.interpolate(pred_mask, size=(h, w), mode="bilinear", align_corners=True)

            """
                mask_cls_results = outputs["pred_logits"]
                mask_pred_results = outputs["pred_masks"]
    
                mask_pred_results = F.interpolate(
                    mask_pred_results,
                    size=(inpurt_tensor.shape[-2], inpurt_tensor.shape[-1]),
                    mode="bilinear",
                    align_corners=False,
                )
                pred_masks = self.semantic_inference(mask_cls_results, mask_pred_results)
                    def semantic_inference(self, mask_cls, mask_pred):
                        mask_cls = F.softmax(mask_cls, dim=-1)[..., 1:]
                        mask_pred = mask_pred.sigmoid()
                        semseg = torch.einsum("bqc,bqhw->bchw", mask_cls, mask_pred)
                        return semseg.cpu().numpy()
                
                mask_img = np.argmax(pred_masks, axis=1)[0]
            """


class MaskFormerHead(nn.Module):
    def __init__(self, input_shape, nclass: int = 2):
        super().__init__()
        self.pixel_decoder = self.pixel_decoder_init(input_shape)
        self.predictor = self.predictor_init(nclass)

    def pixel_decoder_init(self, input_shape):
        common_stride = 4
        transformer_dropout = 0
        transformer_nheads = 8
        transformer_dim_feedforward = 1024
        transformer_enc_layers = 4
        conv_dim = 256
        mask_dim = 256
        transformer_in_features = ["res3", "res4", "res5"]

        pixel_decoder = MSDeformAttnPixelDecoder(input_shape,
                                                 transformer_dropout,
                                                 transformer_nheads,
                                                 transformer_dim_feedforward,
                                                 transformer_enc_layers,
                                                 conv_dim,
                                                 mask_dim,
                                                 transformer_in_features,
                                                 common_stride)
        return pixel_decoder

    def predictor_init(self, nclass):
        in_channels = 256
        num_classes = nclass
        hidden_dim = 256
        num_queries = 100
        nheads = 8
        dim_feedforward = 2048
        dec_layers = 10 - 1
        pre_norm = False
        mask_dim = 256
        enforce_input_project = False
        mask_classification = True
        predictor = MultiScaleMaskedTransformerDecoder(in_channels,
                                                       num_classes,
                                                       mask_classification,
                                                       hidden_dim,
                                                       num_queries,
                                                       nheads,
                                                       dim_feedforward,
                                                       dec_layers,
                                                       pre_norm,
                                                       mask_dim,
                                                       enforce_input_project)
        return predictor

    def forward(self, features, mask=None):
        mask_features, transformer_encoder_features, multi_scale_features = \
            self.pixel_decoder.forward_features(features)
        # print("mask_features", mask_features.shape)
        predictions = self.predictor(multi_scale_features, mask_features, mask)
        # predictions["mask_features"] = mask_features
        return predictions



import os
import numpy as np
from PIL import Image
import cv2
import math
import h5py
import torch
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm

Image.MAX_IMAGE_PIXELS = None


def crop_from_one(image, label, head=(0, 0), size=512):
    x, y = head
    return image[x:x + size, y:y + size, :], label[x:x + size, y:y + size]


def createHDF5(imagePath, labelPath, output_hdf5, downsample, size=512):
    print(f"==============[{downsample}]==============")
    print(imagePath)
    print(labelPath)
    print(output_hdf5)
    ImageNums = len(os.listdir(imagePath))
    oneImage = os.listdir(imagePath)[0]
    H0, W0, _ = cv2.imread(os.path.join(imagePath, oneImage)).shape
    h = round(H0 / downsample)
    w = round(W0 / downsample)
    nums = math.ceil(h / size) * math.ceil(w / size) * ImageNums
    with h5py.File(output_hdf5, "w") as hf:
        image_dataset = hf.create_dataset("images", shape=(nums, size, size, 3), dtype="uint8")
        label_dataset = hf.create_dataset("labels", shape=(nums, size, size), dtype="uint8")
        idx = 0
        for name in tqdm(os.listdir(imagePath)):
            name = name[:-4]
            image0 = cv2.imread(os.path.join(imagePath, name + ".jpg"))
            label0 = np.array(Image.open(os.path.join(labelPath, name + ".png"))).astype(np.uint8)
            image = cv2.resize(image0, (w, h), interpolation=cv2.INTER_LINEAR)
            label = cv2.resize(label0, (w, h), interpolation=cv2.INTER_NEAREST)
            label[label > 0] = 1  # FBP is not need # TODO
            for i in range(0, h, size):
                i = h - size if i + size > h else i
                for j in range(0, w, size):
                    j = w - size if j + size > w else j
                    img, ann = crop_from_one(image=image, label=label, head=(i, j), size=size)
                    image_dataset[idx] = img
                    label_dataset[idx] = ann

                    idx += 1
        hf.close()


class Single_View_HDF5(Dataset):
    def __init__(self, HDF5path):
        super(Single_View_HDF5, self).__init__()
        if not os.path.exists(HDF5path):
            raise FileNotFoundError(f"HDF5 file not found: {HDF5path}")

        if not h5py.is_hdf5(HDF5path):
            raise ValueError(f"Invalid HDF5 file: {HDF5path}")

        hf = h5py.File(HDF5path)
        self.images = hf["images"]
        self.labels = hf["labels"]

    def __len__(self):
        return self.labels.shape[0]

    # def __del__(self):
    #     self.hf.close()

    def __getitem__(self, idx):
        img = self.images[idx]
        ann = self.labels[idx]
        maskTensor = torch.from_numpy(ann).type(torch.LongTensor)
        out = {
            "img": torch.from_numpy(np.transpose(img / 255, (2, 0, 1))).type(torch.FloatTensor),
            "mask": maskTensor
        }
        return out


if __name__ == "__main__":
    # createHDF5(imagePath=r"F:\data_lwc\water\test\img",
    #            labelPath=r"F:\data_lwc\water\test\label",
    #            output_hdf5="water_d5_test.h5", downsample=5, size=512)
    root = r"/scratch/liwanchun"
    path = {"HPD": "HPD",
            "water": "GLH_water",
            "road": "road_data",
            "building": "building_data",
            "FBP": "FBP_data"}
    #
    for data_type in ("building", "water"):  # ("water", "road", "building"):
        for split in ("train", "val"):
            imagePath = os.path.join(root, path[data_type], split, "img")
            labelPath = os.path.join(root, path[data_type], split, "label")
            assert os.path.exists(imagePath) and os.path.exists(labelPath)
            for d in [1]:  # [1, 2, 3, 4, 6, 8]:
                outPath = os.path.join(root, "Large_Size_Datasets", f"{data_type}_{split}_d{d}.h5")
                createHDF5(imagePath=imagePath,
                           labelPath=labelPath,
                           output_hdf5=outPath,
                           downsample=d,
                           size=512)
    # dataset = Single_View_HDF5("water_d5_test.h5")
    # print(len(dataset))
    # batch_size = 2
    # loader = DataLoader(dataset, batch_size=batch_size, shuffle=True, pin_memory=True)
    # # 使用数据加载器进行训练
    # from torchvision.utils import save_image
    #
    # for batch in loader:
    #     img = batch["img"]
    #     print(img.device)
    #     gt = batch["mask"]
    #     save_image(img[0], 'h5img.tif')
    #     save_image(gt[0].type(torch.FloatTensor), 'h5mask.jpg')
    #     print("1")
    #     break
    # for batch in loader:
    #     gt = batch["mask"]
    #     print("2")
    #     break



import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat
from model.semseg.base import BaseNet
from model.semseg.GCM import Global_Context_Module
from timm.models.layers import DropPath, to_2tuple, trunc_normal_
import timm


class ConvBNReLU(nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size=3, dilation=1, stride=1, norm_layer=nn.BatchNorm2d,
                 bias=False):
        super(ConvBNReLU, self).__init__(
            nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, bias=bias,
                      dilation=dilation, stride=stride, padding=((stride - 1) + dilation * (kernel_size - 1)) // 2),
            norm_layer(out_channels),
            nn.ReLU6()
        )


class ConvBN(nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size=3, dilation=1, stride=1, norm_layer=nn.BatchNorm2d,
                 bias=False):
        super(ConvBN, self).__init__(
            nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, bias=bias,
                      dilation=dilation, stride=stride, padding=((stride - 1) + dilation * (kernel_size - 1)) // 2),
            norm_layer(out_channels)
        )


class Conv(nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size=3, dilation=1, stride=1, bias=False):
        super(Conv, self).__init__(
            nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, bias=bias,
                      dilation=dilation, stride=stride, padding=((stride - 1) + dilation * (kernel_size - 1)) // 2)
        )


class SeparableConvBNReLU(nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, dilation=1,
                 norm_layer=nn.BatchNorm2d):
        super(SeparableConvBNReLU, self).__init__(
            nn.Conv2d(in_channels, in_channels, kernel_size, stride=stride, dilation=dilation,
                      padding=((stride - 1) + dilation * (kernel_size - 1)) // 2,
                      groups=in_channels, bias=False),
            norm_layer(out_channels),
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
            nn.ReLU6()
        )


class SeparableConvBN(nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, dilation=1,
                 norm_layer=nn.BatchNorm2d):
        super(SeparableConvBN, self).__init__(
            nn.Conv2d(in_channels, in_channels, kernel_size, stride=stride, dilation=dilation,
                      padding=((stride - 1) + dilation * (kernel_size - 1)) // 2,
                      groups=in_channels, bias=False),
            norm_layer(out_channels),
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        )


class SeparableConv(nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, dilation=1):
        super(SeparableConv, self).__init__(
            nn.Conv2d(in_channels, in_channels, kernel_size, stride=stride, dilation=dilation,
                      padding=((stride - 1) + dilation * (kernel_size - 1)) // 2,
                      groups=in_channels, bias=False),
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        )


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.ReLU6, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Conv2d(in_features, hidden_features, 1, 1, 0, bias=True)
        self.act = act_layer()
        self.fc2 = nn.Conv2d(hidden_features, out_features, 1, 1, 0, bias=True)
        self.drop = nn.Dropout(drop, inplace=True)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class GlobalLocalAttention(nn.Module):
    def __init__(self,
                 dim=256,
                 num_heads=16,
                 qkv_bias=False,
                 window_size=8,
                 relative_pos_embedding=True
                 ):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // self.num_heads
        self.scale = head_dim ** -0.5
        self.ws = window_size

        self.qkv = Conv(dim, 3 * dim, kernel_size=1, bias=qkv_bias)
        self.local1 = ConvBN(dim, dim, kernel_size=3)
        self.local2 = ConvBN(dim, dim, kernel_size=1)
        self.proj = SeparableConvBN(dim, dim, kernel_size=window_size)

        self.attn_x = nn.AvgPool2d(kernel_size=(window_size, 1), stride=1, padding=(window_size // 2 - 1, 0))
        self.attn_y = nn.AvgPool2d(kernel_size=(1, window_size), stride=1, padding=(0, window_size // 2 - 1))

        self.relative_pos_embedding = relative_pos_embedding

        if self.relative_pos_embedding:
            # define a parameter table of relative position bias
            self.relative_position_bias_table = nn.Parameter(
                torch.zeros((2 * window_size - 1) * (2 * window_size - 1), num_heads))  # 2*Wh-1 * 2*Ww-1, nH

            # get pair-wise relative position index for each token inside the window
            coords_h = torch.arange(self.ws)
            coords_w = torch.arange(self.ws)
            coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
            coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
            relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
            relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
            relative_coords[:, :, 0] += self.ws - 1  # shift to start from 0
            relative_coords[:, :, 1] += self.ws - 1
            relative_coords[:, :, 0] *= 2 * self.ws - 1
            relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
            self.register_buffer("relative_position_index", relative_position_index)

            trunc_normal_(self.relative_position_bias_table, std=.02)

    def pad(self, x, ps):
        _, _, H, W = x.size()
        if W % ps != 0:
            x = F.pad(x, (0, ps - W % ps), mode='reflect')
        if H % ps != 0:
            x = F.pad(x, (0, 0, 0, ps - H % ps), mode='reflect')
        return x

    def pad_out(self, x):
        x = F.pad(x, pad=(0, 1, 0, 1), mode='reflect')
        return x

    def forward(self, x):
        B, C, H, W = x.shape

        local = self.local2(x) + self.local1(x)

        x = self.pad(x, self.ws)
        B, C, Hp, Wp = x.shape
        qkv = self.qkv(x)

        q, k, v = rearrange(qkv, 'b (qkv h d) (hh ws1) (ww ws2) -> qkv (b hh ww) h (ws1 ws2) d', h=self.num_heads,
                            d=C // self.num_heads, hh=Hp // self.ws, ww=Wp // self.ws, qkv=3, ws1=self.ws, ws2=self.ws)

        dots = (q @ k.transpose(-2, -1)) * self.scale

        if self.relative_pos_embedding:
            relative_position_bias = self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
                self.ws * self.ws, self.ws * self.ws, -1)  # Wh*Ww,Wh*Ww,nH
            relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww
            dots += relative_position_bias.unsqueeze(0)

        attn = dots.softmax(dim=-1)
        attn = attn @ v

        attn = rearrange(attn, '(b hh ww) h (ws1 ws2) d -> b (h d) (hh ws1) (ww ws2)', h=self.num_heads,
                         d=C // self.num_heads, hh=Hp // self.ws, ww=Wp // self.ws, ws1=self.ws, ws2=self.ws)

        attn = attn[:, :, :H, :W]

        out = self.attn_x(F.pad(attn, pad=(0, 0, 0, 1), mode='reflect')) + \
              self.attn_y(F.pad(attn, pad=(0, 1, 0, 0), mode='reflect'))

        out = out + local
        out = self.pad_out(out)
        out = self.proj(out)
        # print(out.size())
        out = out[:, :, :H, :W]

        return out


class Block(nn.Module):
    def __init__(self, dim=256, num_heads=16, mlp_ratio=4., qkv_bias=False, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.ReLU6, norm_layer=nn.BatchNorm2d, window_size=8):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = GlobalLocalAttention(dim, num_heads=num_heads, qkv_bias=qkv_bias, window_size=window_size)

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, out_features=dim, act_layer=act_layer,
                       drop=drop)
        self.norm2 = norm_layer(dim)

    def forward(self, x):
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class WF(nn.Module):
    def __init__(self, in_channels=128, decode_channels=128, eps=1e-8):
        super(WF, self).__init__()
        self.pre_conv = Conv(in_channels, decode_channels, kernel_size=1)

        self.weights = nn.Parameter(torch.ones(2, dtype=torch.float32), requires_grad=True)
        self.eps = eps
        self.post_conv = ConvBNReLU(decode_channels, decode_channels, kernel_size=3)

    def forward(self, x, res):
        x = F.interpolate(x, scale_factor=2, mode='bilinear', align_corners=False)
        weights = nn.ReLU()(self.weights)
        fuse_weights = weights / (torch.sum(weights, dim=0) + self.eps)
        x = fuse_weights[0] * self.pre_conv(res) + fuse_weights[1] * x
        x = self.post_conv(x)
        return x


class FeatureRefinementHead(nn.Module):
    def __init__(self, in_channels=64, decode_channels=64):
        super().__init__()
        self.pre_conv = Conv(in_channels, decode_channels, kernel_size=1)

        self.weights = nn.Parameter(torch.ones(2, dtype=torch.float32), requires_grad=True)
        self.eps = 1e-8
        self.post_conv = ConvBNReLU(decode_channels, decode_channels, kernel_size=3)

        self.pa = nn.Sequential(
            nn.Conv2d(decode_channels, decode_channels, kernel_size=3, padding=1, groups=decode_channels),
            nn.Sigmoid())
        self.ca = nn.Sequential(nn.AdaptiveAvgPool2d(1),
                                Conv(decode_channels, decode_channels // 16, kernel_size=1),
                                nn.ReLU6(),
                                Conv(decode_channels // 16, decode_channels, kernel_size=1),
                                nn.Sigmoid())

        self.shortcut = ConvBN(decode_channels, decode_channels, kernel_size=1)
        self.proj = SeparableConvBN(decode_channels, decode_channels, kernel_size=3)
        self.act = nn.ReLU6()

    def forward(self, x, res):
        x = F.interpolate(x, scale_factor=2, mode='bilinear', align_corners=False)
        weights = nn.ReLU()(self.weights)
        fuse_weights = weights / (torch.sum(weights, dim=0) + self.eps)
        x = fuse_weights[0] * self.pre_conv(res) + fuse_weights[1] * x
        x = self.post_conv(x)
        shortcut = self.shortcut(x)
        pa = self.pa(x) * x
        ca = self.ca(x) * x
        x = pa + ca
        x = self.proj(x) + shortcut
        x = self.act(x)

        return x


class AuxHead(nn.Module):

    def __init__(self, in_channels=64, num_classes=8):
        super().__init__()
        self.conv = ConvBNReLU(in_channels, in_channels)
        self.drop = nn.Dropout(0.1)
        self.conv_out = Conv(in_channels, num_classes, kernel_size=1)

    def forward(self, x, h, w):
        feat = self.conv(x)
        feat = self.drop(feat)
        feat = self.conv_out(feat)
        feat = F.interpolate(feat, size=(h, w), mode='bilinear', align_corners=False)
        return feat


class Decoder(nn.Module):
    def __init__(self,
                 encoder_channels=(64, 128, 256, 512),
                 decode_channels=64,
                 dropout=0.1,
                 window_size=8,
                 num_classes=6):
        super(Decoder, self).__init__()

        self.pre_conv = ConvBN(encoder_channels[-1], decode_channels, kernel_size=1)
        self.b4 = Block(dim=decode_channels, num_heads=8, window_size=window_size)

        self.b3 = Block(dim=decode_channels, num_heads=8, window_size=window_size)
        self.p3 = WF(encoder_channels[-2], decode_channels)

        self.b2 = Block(dim=decode_channels, num_heads=8, window_size=window_size)
        self.p2 = WF(encoder_channels[-3], decode_channels)

        if self.training:
            self.up4 = nn.UpsamplingBilinear2d(scale_factor=4)
            self.up3 = nn.UpsamplingBilinear2d(scale_factor=2)
            self.aux_head = AuxHead(decode_channels, num_classes)

        self.p1 = FeatureRefinementHead(encoder_channels[-4], decode_channels)

        self.segmentation_head = nn.Sequential(ConvBNReLU(decode_channels, decode_channels),
                                               nn.Dropout2d(p=dropout, inplace=True),
                                               Conv(decode_channels, num_classes, kernel_size=1))
        self.init_weight()

    def forward(self, res1, res2, res3, res4, h, w):
        if self.training:
            x = self.b4(self.pre_conv(res4))
            h4 = self.up4(x)

            x = self.p3(x, res3)
            x = self.b3(x)
            h3 = self.up3(x)

            x = self.p2(x, res2)
            x = self.b2(x)
            h2 = x
            x = self.p1(x, res1)
            x = self.segmentation_head(x)
            x = F.interpolate(x, size=(h, w), mode='bilinear', align_corners=False)

            ah = h4 + h3 + h2
            ah = self.aux_head(ah, h, w)  # TODO deleted

            return x, ah
        else:
            x = self.b4(self.pre_conv(res4))
            x = self.p3(x, res3)
            x = self.b3(x)

            x = self.p2(x, res2)
            x = self.b2(x)

            x = self.p1(x, res1)

            x = self.segmentation_head(x)
            x = F.interpolate(x, size=(h, w), mode='bilinear', align_corners=False)

            return x

    def init_weight(self):
        for m in self.children():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, a=1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)


class UNetFormer(BaseNet):
    def __init__(self, backbone,
                 nclass,
                 decode_channels=64,
                 dropout=0.1,
                 window_size=8,
                 isContext: bool = False):
        super(UNetFormer, self).__init__(backbone)
        # TODO
        self.isContext = isContext
        print(f"isContext = {self.isContext}")
        if self.isContext:
            self.GCM = Global_Context_Module(channels=self.backbone.channels)
        self.decoder = Decoder(self.backbone.channels, decode_channels, dropout, window_size, nclass)

    def base_forward0(self, x):
        h, w = x.size()[-2:]
        res1, res2, res3, res4 = self.backbone(x)
        if self.training:
            x, ah = self.decoder(res1, res2, res3, res4, h, w)
            # return x, ah
        else:
            x = self.decoder(res1, res2, res3, res4, h, w)
        return x

    def base_forward_context(self, xs):
        x1, x2, x3 = xs
        with torch.no_grad():
            f2 = self.backbone.base_forward(x2)
            f3 = self.backbone.base_forward(x3)
        f1 = self.backbone.base_forward(x1)
        h, w = x1.size()[-2:]
        res1, res2, res3, res4 = self.GCM(f1, f2, f3)
        if self.training:
            x, ah = self.decoder(res1, res2, res3, res4, h, w)
            # return x, ah
        else:
            x = self.decoder(res1, res2, res3, res4, h, w)
        return x

    def base_forward(self, x):
        if self.isContext:
            return self.base_forward_context(x)
        else:
            return self.base_forward0(x)


class UNetFormer0(nn.Module):
    def __init__(self,
                 decode_channels=64,
                 dropout=0.1,
                 backbone_name='swsl_resnet18',
                 pretrained=True,
                 window_size=8,
                 num_classes=6
                 ):
        super().__init__()

        self.backbone = timm.create_model(backbone_name, features_only=True, output_stride=32,
                                          out_indices=(1, 2, 3, 4), pretrained=pretrained)
        encoder_channels = self.backbone.feature_info.channels()

        self.decoder = Decoder(encoder_channels, decode_channels, dropout, window_size, num_classes)

    def forward(self, x):
        h, w = x.size()[-2:]
        res1, res2, res3, res4 = self.backbone(x)
        if self.training:
            x, ah = self.decoder(res1, res2, res3, res4, h, w)
            return x, ah
        else:
            x = self.decoder(res1, res2, res3, res4, h, w)
            return x



import os
import cv2
import h5py
import random
import numpy as np
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data.dataset import Dataset
from torch.utils.data import DataLoader
from torch.cuda.amp import autocast, GradScaler

from utils.metrics import Evaluator
from utils.net_utils import loadWeight, EarlyStopping, checkFile, seed_torch, count_trainable_params
from build import *
from Fed import FedTrain
from utils.path import checkPath, photo_metric_distortion, normalize_function
# TODO
from configs.args1024 import settings

# 设置随机种子
seed_torch()
os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "max_split_size_mb:4"

# 定义加载网络模型的函数
def loadNet(args: settings):
    net = build_model(args)
    pth_best = os.path.join(args.pth_path, f"{args.pth_name}_best.pth")
    assert os.path.isfile(pth_best)
    loadWeight(path=pth_best, net=net)
    net.eval()
    return net

# 自定义数据集类，用于处理HDF5格式的数据
class Views_HDF5(Dataset):
    def __init__(self, HDF5path, training=False):
        super(Views_HDF5, self).__init__()
        if not os.path.exists(HDF5path):
            raise FileNotFoundError(f"HDF5 file not found: {HDF5path}")

        if not h5py.is_hdf5(HDF5path):
            raise ValueError(f"Invalid HDF5 file: {HDF5path}")

        hf = h5py.File(HDF5path)
        self.images = hf["images"]
        self.labels = hf["labels"]
        self.training = training

    def __len__(self):
        return self.labels.shape[0]

    def __getitem__(self, idx):
        img = self.images[idx]
        ann = self.labels[idx]
        if self.training:
            img, ann = self.aug(img, ann)  # 数据增强
        maskTensor = torch.from_numpy(ann).type(torch.LongTensor)
        out = {
            "img": normalize_function(img),
            "mask": maskTensor
        }
        return out

    @staticmethod
    def aug(img, ann):
        def color_distortion(image):
            # 随机改变图像的亮度、对比度和饱和度
            alpha = 1.0 + np.random.uniform(-0.5, 0.5)  # 亮度
            beta = 0.5 + np.random.uniform(-0.5, 0.5)  # 对比度
            saturation_scale = 0.5 + np.random.uniform(0, 1)  # 饱和度

            image = cv2.convertScaleAbs(image, alpha=alpha, beta=beta)
            hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

            # 调整饱和度
            hsv[:, :, 1] = np.clip(hsv[:, :, 1] * saturation_scale, 0, 255)

            distorted_image = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
            return distorted_image

        def gaussian_noise(image):
            row, col, ch = image.shape
            mean = 0
            var = 0.1
            sigma = var ** 0.5
            gauss = np.random.normal(mean, sigma, (row, col, ch))
            noisy = np.clip(image + gauss * 255, 0, 255).astype(np.uint8)
            return noisy

        def random_cover(image):
            row, col, ch = image.shape
            r = random.randint(8, row // 8)
            c = random.randint(8, col // 8)
            m = np.zeros((r, c, ch))
            sr = random.randint(0, row - r - 1)
            sc = random.randint(0, col - c - 1)
            image[sr:sr + r, sc:sc + c, :] = m
            return image

        if random.random() < 0.5:
            flip_type = random.choice([0, 1])
            img = cv2.flip(img, flip_type)
            ann = cv2.flip(ann, flip_type)
        
        return img, ann

# 训练函数
def train(data):
    size = 512

    # 设置GPU
    print(torch.cuda.device_count())
    assert torch.cuda.device_count() > 0
    device = torch.device("cuda:0")

    d1 = 1
    print(f"==============[{data}]==============")

    # 保存路径？ 参数设置
    if data == "FBP":
        pth_root = rf"./weight/FBP_normal_sswinl_upernet{size}_{d1}"
        args = settings(segHead="upernet", downsample=d1, backbone="sswin_l", isContext=False)
    elif data == "water":
        pth_root = rf"./weight/water_normal_sswinl_upernet_{size}_{d1}"
        args = settings(segHead="upernet", data_type="water", downsample=d1, backbone="sswin_l", isContext=False)
    elif data == "building":
        pth_root = rf"./weight/building_normal_sswinl_upernet_{size}_{d1}"
        args = settings(segHead="upernet", data_type="building", downsample=d1, backbone="sswin_l", isContext=False)
    else:
        raise ValueError

    # 加载训练集和验证集
    checkPath(path=pth_root)
    train_data = Views_HDF5(HDF5path=os.path.join("/scratch/liwanchun/Large_Size_Datasets",
                                                  f"{data}_train_d{d1}.h5"),
                            training=True)
    train_loader = DataLoader(train_data,
                              batch_size=8 if size == 512 else 1,
                              num_workers=2,
                              shuffle=True,
                              drop_last=True)
    val_data = Views_HDF5(HDF5path=os.path.join("/scratch/liwanchun/Large_Size_Datasets",
                                                f"{data}_val_d{d1}.h5"))
    val_loader = DataLoader(val_data,
                            batch_size=1,
                            shuffle=True,
                            drop_last=True)

    # 模型加载
    net1 = build_model(args=args)
    path = os.path.join(pth_root, "best.pth")
    if os.path.isfile(path):
        print(f"Loading net from : {path}")
        check = torch.load(path)
        net1.load_state_dict(check)
    net1.to(device)

    # 优化器
    optimizer, scheduler = build_optimizer(net=net1, args=args)
    evaluator, earlystopping = build_evaluator(args)
    earlystopping.criterion = 0

    # 创建损失函数
    criterion = build_loss(args)

    # 训练过程
    max_epochs = 80
    print(f"max epochs:{max_epochs}")
    torch.save(net1.state_dict(), os.path.join(pth_root, f"0.pth"))

    for epoch in range(1, max_epochs + 1):
        print(f"##########################[{epoch}]##########################")
        print(f"Training [{epoch}]:")
        # training
        net1.train()
        print("learn rate = {:.9f}".format(optimizer.param_groups[0]['lr']))
        print("learn rate = {:.9f}".format(optimizer.param_groups[-1]['lr']))
        bar_train = tqdm(train_loader, desc=f'Train::', position=0, leave=True, mininterval=60)

        # 获取图像和标签，进行训练
        for idx, batch in enumerate(bar_train):
            torch.cuda.empty_cache()
            img1 = batch["img"].to(device)
            mask = batch["mask"].to(device)
            torch.cuda.empty_cache()

            out = net1(img1)
            loss = criterion(out, mask)
            net1.zero_grad()
            (loss["loss_tot"]).backward()
            optimizer.step()

            #每2两个batch 更新进度条和统计信息
            if idx % 2 == 0:
                postfix_args = {k: "{:.5f}".format(v.item()) for k, v in loss.items() \
                                if k in ("loss_tot", "loss_ce", "loss_mask", "loss_dice")}
                bar_train.set_postfix(**postfix_args)
                bar_train.update()

        #保存当前的权重
        torch.save(net1.state_dict(), os.path.join(pth_root, f"{epoch}.pth"))
        print(f"{epoch}.pth was saved to {pth_root}")
        if checkFile(path=os.path.join(pth_root, f"{epoch - 1}.pth")):
            os.remove(os.path.join(pth_root, f"{epoch - 1}.pth"))

        # 验证阶段
        net1.eval()
        print(f"Validation [{epoch}]:")

        torch.cuda.empty_cache()
        bar_val = tqdm(val_loader, desc=f'Val::', position=0, leave=True, mininterval=60)
        evaluator.reset()
        with torch.no_grad():
            for idx, batch in enumerate(bar_val):
                img1 = batch["img"].to(device)
                outputs = net1(img1)
                gt = batch["mask"].to(device)
                # pred = outputs[:, 1:, :, :]  # TODO FBP
                pred = outputs
                pre = np.argmax(pred.cpu().numpy(), axis=1)[0]
                gt = gt.squeeze(0).cpu().numpy()

                evaluator.add_batch(gt, pre)
                if idx % 2 == 0:
                    postfix_args1 = {"id": f"{idx}"}
                    bar_val.set_postfix(**postfix_args1)
                    bar_val.update()

        MIoU = evaluator.Mean_Intersection_over_Union()
        print(f"MIoU = {MIoU}")
        if earlystopping.CheckStopping(new_criterion=MIoU):
            torch.save(net1.state_dict(), os.path.join(pth_root, "best.pth"))
            print("BETTER MODEL SAVED")
        scheduler.step(earlystopping.criterion)

        print(f"best_MIoU = {earlystopping.criterion}")

        if earlystopping.EarlyStopping:
            print("EarlyStooping")
            return

def select_data(data_type, batch_size):

    train_data_path = "/scratch/liwanchun/Large_Size_Datasets"
    val_data_path = "/scratch/liwanchun/Large_Size_Datasets"
    size = 512
    d1 = 1
    # 根据数据类型加载训练集和验证集
    if data_type == "FBP":
        train_dataset_path = os.path.join(train_data_path, f"FBP_train_d{d1}.h5")
        val_dataset_path = os.path.join(val_data_path, f"FBP_val_d{d1}.h5")
        pth_root = rf"./weight/FBP_normal_sswinl_upernet{size}_{d1}"
        args = settings(downsample=d1, isContext=False)
    elif data_type == "water":
        train_dataset_path = os.path.join(train_data_path, f"water_train_d{d1}.h5")
        val_dataset_path = os.path.join(val_data_path, f"water_val_d{d1}.h5")
        pth_root = rf"./weight/water_normal_sswinl_upernet_{size}_{d1}"
        args = settings(data_type="water", downsample=d1, isContext=False)
    elif data_type == "building":
        train_dataset_path = os.path.join(train_data_path, f"building_train_d{d1}.h5")
        val_dataset_path = os.path.join(val_data_path, f"building_val_d{d1}.h5")
        pth_root = rf"./weight/building_normal_sswinl_upernet_{size}_{d1}"
        args = settings(data_type="building", downsample=d1, isContext=False)
    else:
        raise ValueError("Invalid data type specified")

    train_data = Views_HDF5(HDF5path=train_dataset_path, training=True)
    val_data = Views_HDF5(HDF5path=val_dataset_path)

    #train_loader = DataLoader(train_data, batch_size=8 if size == 512 else 1, num_workers=2, shuffle=True, drop_last=True)
    train_loader = DataLoader(train_data, batch_size=batch_size, num_workers=2, shuffle=True, drop_last=True)
    val_loader = DataLoader(val_data, batch_size=1, shuffle=True, drop_last=True)


    return train_loader, val_loader, args, pth_root

def design_model(decoder, encoder, args, pth_root):

    print(torch.cuda.device_count())
    assert torch.cuda.device_count() > 0
    device = torch.device("cuda:0")
    # 根据给定参数创建模型
    #args = settings(segHead=decoder, backbone=encoder)
    args.segHead = decoder
    args.backbone = encoder
    model = build_model(args=args)
    path = os.path.join(pth_root, "best.pth")
    if os.path.isfile(path):
        print(f"Loading net from : {path}")
        check = torch.load(path)
        model.load_state_dict(check)
    model.to(device)
    return model,args,device

def design_optimizer(model, args):
    optimizer, scheduler = build_optimizer(net=model, args=args)
    evaluator, earlystopping = build_evaluator(args)
    earlystopping.criterion = 0
    return optimizer,scheduler,evaluator,earlystopping

def design_loss(args):
    criterion = build_loss(args)
    return criterion

def train_model(epoch, model, pth_root, train_loader, val_loader, device,optimizer, scheduler, evaluator, earlystopping, criterion):
    # 训练过程
    max_epochs = epoch
    net1 = model

    print(f"max epochs:{max_epochs}")
    torch.save(net1.state_dict(), os.path.join(pth_root, f"0.pth"))

    for epoch in range(1, max_epochs + 1):
        print(f"##########################[{epoch}]##########################")
        print(f"Training [{epoch}]:")
        # training
        net1.train()
        print("learn rate = {:.9f}".format(optimizer.param_groups[0]['lr']))
        print("learn rate = {:.9f}".format(optimizer.param_groups[-1]['lr']))
        bar_train = tqdm(train_loader, desc=f'Train::', position=0, leave=True, mininterval=60)

        # 获取图像和标签，进行训练
        for idx, batch in enumerate(bar_train):
            torch.cuda.empty_cache()
            img1 = batch["img"].to(device)
            mask = batch["mask"].to(device)
            torch.cuda.empty_cache()

            out = net1(img1)
            loss = criterion(out, mask)
            net1.zero_grad()
            (loss["loss_tot"]).backward()
            optimizer.step()

            # 每2两个batch 更新进度条和统计信息
            if idx % 2 == 0:
                postfix_args = {k: "{:.5f}".format(v.item()) for k, v in loss.items() \
                                if k in ("loss_tot", "loss_ce", "loss_mask", "loss_dice")}
                bar_train.set_postfix(**postfix_args)
                bar_train.update()

        # 保存当前的权重
        torch.save(net1.state_dict(), os.path.join(pth_root, f"{epoch}.pth"))
        print(f"{epoch}.pth was saved to {pth_root}")
        if checkFile(path=os.path.join(pth_root, f"{epoch - 1}.pth")):
            os.remove(os.path.join(pth_root, f"{epoch - 1}.pth"))

        # 验证阶段
        net1.eval()
        print(f"Validation [{epoch}]:")

        torch.cuda.empty_cache()
        bar_val = tqdm(val_loader, desc=f'Val::', position=0, leave=True, mininterval=60)
        evaluator.reset()
        with torch.no_grad():
            for idx, batch in enumerate(bar_val):
                img1 = batch["img"].to(device)
                outputs = net1(img1)
                gt = batch["mask"].to(device)
                # pred = outputs[:, 1:, :, :]  # TODO FBP
                pred = outputs
                pre = np.argmax(pred.cpu().numpy(), axis=1)[0]
                gt = gt.squeeze(0).cpu().numpy()

                evaluator.add_batch(gt, pre)
                if idx % 2 == 0:
                    postfix_args1 = {"id": f"{idx}"}
                    bar_val.set_postfix(**postfix_args1)
                    bar_val.update()

        MIoU = evaluator.Mean_Intersection_over_Union()
        print(f"MIoU = {MIoU}")
        if earlystopping.CheckStopping(new_criterion=MIoU):
            torch.save(net1.state_dict(), os.path.join(pth_root, "best.pth"))
            print("BETTER MODEL SAVED")
        scheduler.step(earlystopping.criterion)

        print(f"best_MIoU = {earlystopping.criterion}")

        if earlystopping.EarlyStopping:
            print("EarlyStooping")
            return

# if __name__ == "__main__":
#     # import argparse
#     #
#     # parser = argparse.ArgumentParser('Set PFNetPlus Train', add_help=False)
#     # parser.add_argument('--data', type=str, default="water")
#     #
#     # args = parser.parse_args()
#     # print(f"==============[{args.data}]==============")
#     # train(data=args.data)
#
#     #数据选择
#     train_loader, val_loader, args, pth_root = select_data("FBP", 8)
#
#     #模型设计
#     model, args, device = design_model("upernet", "sswin-l", args, pth_root)
#
#     #优化器设计
#     optimizer, scheduler, evaluator, earlystopping = design_optimizer(model, args)
#
#     #损失函数设计
#     criterion = design_loss(args)
#
#     #模型训练
#     train_model(80, model, pth_root, train_loader,val_loader, device, optimizer, scheduler, evaluator, earlystopping, criterion)



# Copyright (c) Facebook, Inc. and its affiliates.
# Modified by Bowen Cheng from https://github.com/facebookresearch/detr/blob/master/models/detr.py
"""
MaskFormer criterion.
"""
from typing import List, Optional
import torch
from torch import Tensor
import numpy as np
import torch.nn.functional as F
import torch.distributed as dist
from torch import nn
import sys
import os
import torchvision

sys.path.append(os.path.dirname(__file__) + os.sep + '../')

# 计算二维列表每个列的最大值
def _max_by_axis(the_list):
    # type: (List[List[int]]) -> List[int]
    maxes = the_list[0]
    for sublist in the_list[1:]:
        for index, item in enumerate(sublist):
            maxes[index] = max(maxes[index], item)
    return maxes

# 获取分布式训练的进程数量
def get_world_size() -> int:
    if not dist.is_available():
        return 1
    if not dist.is_initialized():
        return 1
    return dist.get_world_size()


def reduce_dict(input_dict, average=True):
    """
    归约分布式训练中的字典，将所有进程的字典数据汇总
    """
    world_size = get_world_size()
    if world_size < 2:
        return input_dict
    with torch.no_grad():
        names = []
        values = []
        # 按键排序，确保跨进程一致
        for k in sorted(input_dict.keys()):
            names.append(k)
            values.append(input_dict[k])
        values = torch.stack(values, dim=0)
        dist.all_reduce(values)
        if average:
            values /= world_size
        reduced_dict = {k: v for k, v in zip(names, values)}
    return reduced_dict


class NestedTensor(object):
    def __init__(self, tensors, mask: Optional[Tensor]):
        self.tensors = tensors
        self.mask = mask

    def to(self, device):
        # type: (Device) -> NestedTensor # noqa
        cast_tensor = self.tensors.to(device)
        mask = self.mask
        if mask is not None:
            assert mask is not None
            cast_mask = mask.to(device)
        else:
            cast_mask = None
        return NestedTensor(cast_tensor, cast_mask)

    def decompose(self):
        return self.tensors, self.mask

    def __repr__(self):
        return str(self.tensors)

# 将tensor列表转换为NestedTensor
def nested_tensor_from_tensor_list(tensor_list: List[Tensor]):
    # TODO make this more general
    if tensor_list[0].ndim == 3:
        if torchvision._is_tracing():
            # nested_tensor_from_tensor_list() does not export well to ONNX
            # call _onnx_nested_tensor_from_tensor_list() instead
            return _onnx_nested_tensor_from_tensor_list(tensor_list)

        # TODO make it support different-sized images
        max_size = _max_by_axis([list(img.shape) for img in tensor_list])
        # min_size = tuple(min(s) for s in zip(*[img.shape for img in tensor_list]))
        batch_shape = [len(tensor_list)] + max_size
        b, c, h, w = batch_shape
        dtype = tensor_list[0].dtype
        device = tensor_list[0].device
        tensor = torch.zeros(batch_shape, dtype=dtype, device=device)
        mask = torch.ones((b, h, w), dtype=torch.bool, device=device)
        for img, pad_img, m in zip(tensor_list, tensor, mask):
            pad_img[: img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)
            m[: img.shape[1], : img.shape[2]] = False
    else:
        raise ValueError("not supported")
    return NestedTensor(tensor, mask)


# 处理ONNX tracing中不支持的操作
@torch.jit.unused
def _onnx_nested_tensor_from_tensor_list(tensor_list: List[Tensor]) -> NestedTensor:
    max_size = []
    for i in range(tensor_list[0].dim()):
        max_size_i = torch.max(
            torch.stack([img.shape[i] for img in tensor_list]).to(torch.float32)
        ).to(torch.int64)
        max_size.append(max_size_i)
    max_size = tuple(max_size)

    # work around for
    # pad_img[: img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)
    # m[: img.shape[1], :img.shape[2]] = False
    # which is not yet supported in onnx
    padded_imgs = []
    padded_masks = []
    for img in tensor_list:
        padding = [(s1 - s2) for s1, s2 in zip(max_size, tuple(img.shape))]
        padded_img = torch.nn.functional.pad(img, (0, padding[2], 0, padding[1], 0, padding[0]))
        padded_imgs.append(padded_img)

        m = torch.zeros_like(img[0], dtype=torch.int, device=img.device)
        padded_mask = torch.nn.functional.pad(m, (0, padding[2], 0, padding[1]), "constant", 1)
        padded_masks.append(padded_mask.to(torch.bool))

    tensor = torch.stack(padded_imgs)
    mask = torch.stack(padded_masks)

    return NestedTensor(tensor, mask=mask)


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def dice_loss(
        inputs: torch.Tensor,
        targets: torch.Tensor,
        num_masks: float,
):
    """
    计算DICE损失，类似于用于掩码的广义IOU
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
    """
    inputs = inputs.sigmoid()
    inputs = inputs.flatten(1)
    numerator = 2 * (inputs * targets).sum(-1)
    denominator = inputs.sum(-1) + targets.sum(-1)
    loss = 1 - (numerator + 1) / (denominator + 1)
    return loss.sum() / num_masks


def sigmoid_ce_loss(
        inputs: torch.Tensor,
        targets: torch.Tensor,
        num_masks: float,
):
    """
    计算二值交叉熵损失
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
    Returns:
        Loss tensor
    """
    loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")
    return loss.mean(1).sum() / num_masks


def sigmoid_focal_loss(inputs, targets, num_masks, alpha: float = 0.25, gamma: float = 2):
    """
    RetinaNet中用于密集检测的损失函数: https://arxiv.org/abs/1708.02002.
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
        alpha: (optional) Weighting factor in range (0,1) to balance
                positive vs negative examples. Default = -1 (no weighting).
        gamma: Exponent of the modulating factor (1 - p_t) to
               balance easy vs hard examples.
    Returns:
        Loss tensor
    """
    prob = inputs.sigmoid()
    ce_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")
    p_t = prob * targets + (1 - prob) * (1 - targets)
    loss = ce_loss * ((1 - p_t) ** gamma)

    if alpha >= 0:
        alpha_t = alpha * targets + (1 - alpha) * (1 - targets)
        loss = alpha_t * loss

    return loss.mean(1).sum() / num_masks


def calculate_uncertainty(logits):
    """
    我们通过计算 'logits' 中前景类的预测与 0.0 的 L1 距离来估计不确定性。
    参数:
        logits (Tensor): 形状为 (R, 1, ...) 的张量，代表特定类别或类别无关的预测结果，其中 R 是所有图片中预测的掩码总数，C 是前景类别的数量。值为 logits。
    返回:
        scores (Tensor): 形状为 (R, 1, ...) 的张量，包含不确定性分数，最不确定的位置具有最高的不确定性分数。
    """
    assert logits.shape[1] == 1
    gt_class_logits = logits.clone()
    return -(torch.abs(gt_class_logits))


class SetCriterion(nn.Module):
    """
    这个类用于计算 DETR 的损失。
    过程分为两个步骤：
        1) 计算真实框与模型输出之间的匈牙利匹配。
        2) 对每对匹配的真实值/预测值进行监督（监督类别和框）。
    """

    def __init__(self, num_classes, matcher, weight_dict, eos_coef, losses,
                 num_points, oversample_ratio, importance_sample_ratio, device):
        """
        创建损失计算标准。
        参数:
            num_classes: 对象类别数量，不包括特殊的无对象类别。
            matcher: 能够计算目标和提议之间匹配的模块。
            weight_dict: 包含损失名称为键，及其相对权重为值的字典。
            eos_coef: 应用于无对象类别的相对分类权重。
            losses: 所有将应用的损失的列表。请参阅 get_loss 以获取可用损失的列表。
        """
        super().__init__()
        self.num_classes = num_classes
        self.matcher = matcher
        self.weight_dict = weight_dict
        self.eos_coef = eos_coef
        self.losses = losses
        self.device = device
        empty_weight = torch.ones(self.num_classes + 1).to(device)
        empty_weight[0] = 0  # TODO for FBP
        empty_weight[-1] = self.eos_coef
        self.empty_weight = empty_weight

        # pointwise mask loss parameters
        self.num_points = num_points
        self.oversample_ratio = oversample_ratio
        self.importance_sample_ratio = importance_sample_ratio

    def loss_labels(self, outputs, targets, indices, num_masks):
        """
        分类损失 (NLL)
        目标字典必须包含键 "labels"，其包含维度为 [nb_target_boxes] 的张量。
        """
        assert "pred_logits" in outputs
        src_logits = outputs["pred_logits"].float()

        idx = self._get_src_permutation_idx(indices)
        target_classes_o = torch.cat([t["labels"][J] for t, (_, J) in zip(targets, indices)]).to(self.device)
        target_classes = torch.full(src_logits.shape[:2], 0, dtype=torch.int64, device=src_logits.device)
        target_classes[idx] = target_classes_o

        loss_ce = F.cross_entropy(src_logits.transpose(1, 2), target_classes, self.empty_weight)
        losses = {"loss_ce": loss_ce}
        return losses

    def loss_masks(self, outputs, targets, indices, num_masks):
        """
        计算与掩码相关的损失：焦点损失和 dice 损失。
        目标字典必须包含键 "masks"，其包含维度为 [nb_target_boxes, h, w] 的张量。
        """
        assert "pred_masks" in outputs

        src_idx = self._get_src_permutation_idx(indices)
        tgt_idx = self._get_tgt_permutation_idx(indices)
        src_masks = outputs["pred_masks"]  #
        src_masks = src_masks[src_idx]
        masks = [t["masks"] for t in targets]
        # TODO use valid to mask invalid areas due to padding in loss
        target_masks, valid = nested_tensor_from_tensor_list(masks).decompose()
        # print([x.shape for x in masks])
        # target_masks = torch.cat(masks, dim=0)
        target_masks = target_masks.to(src_masks)
        target_masks = target_masks[tgt_idx]

        # ===================================================================================
        # seg task no need to sample #
        # No need to upsample predictions as we are using normalized coordinates :)
        # N x 1 x H x W

        # src_masks = src_masks[:, None]
        # target_masks = target_masks[:, None]

        # with torch.no_grad():
        #     # sample point_coords
        #     point_coords = get_uncertain_point_coords_with_randomness(
        #         src_masks,
        #         lambda logits: calculate_uncertainty(logits),
        #         self.num_points,
        #         self.oversample_ratio,
        #         self.importance_sample_ratio,
        #     )
        #     # get gt labels
        #     point_labels = point_sample(
        #         target_masks,
        #         point_coords,
        #         align_corners=False,
        #     ).squeeze(1)

        # point_logits = point_sample(
        #     src_masks,
        #     point_coords,
        #     align_corners=False,
        # ).squeeze(1)
        # ===================================================================================
        point_logits = src_masks.flatten(1)
        point_labels = target_masks.flatten(1)

        losses = {
            "loss_mask": sigmoid_ce_loss(point_logits, point_labels, num_masks),
            # sigmoid_focal_loss(point_logits, point_labels, num_masks), #
            "loss_dice": dice_loss(point_logits, point_labels, num_masks)
        }

        del src_masks
        del target_masks
        return losses

    def _get_src_permutation_idx(self, indices):
        # 根据索引排列预测值
        batch_idx = torch.cat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.cat([src for (src, _) in indices])
        return batch_idx, src_idx

    def _get_tgt_permutation_idx(self, indices):
        # 根据索引排列目标值
        batch_idx = torch.cat([torch.full_like(tgt, i) for i, (_, tgt) in enumerate(indices)])
        tgt_idx = torch.cat([tgt for (_, tgt) in indices])
        return batch_idx, tgt_idx

    def get_loss(self, loss, outputs, targets, indices, num_masks):
        loss_map = {
            'labels': self.loss_labels,
            'masks': self.loss_masks,
        }
        assert loss in loss_map, f"do you really want to compute {loss} loss?"
        return loss_map[loss](outputs, targets, indices, num_masks)

    @staticmethod
    def postOutput(outputs):
        for k, v in outputs.items():
            if k == "aux_outputs":
                for i in range(len(v)):
                    outputs[k][i]["pred_masks"] = F.interpolate(
                        outputs[k][i]["pred_masks"].float(),
                        scale_factor=4,
                        mode="bilinear",
                        align_corners=False,
                    )
            elif k == "pred_masks":
                outputs[k] = F.interpolate(
                    outputs[k].float(),
                    scale_factor=4,
                    mode="bilinear",
                    align_corners=False,
                )
        return outputs

    def forward(self, outputs, gt_masks):
        """
        执行损失计算。
        参数:
             outputs: 张量的字典，参见模型的输出规范以获取格式。
             gt_masks: [bs, h_net_output, w_net_output]
        """
        # H, W = outputs["pred_masks"].shape[2:]
        # gt_masks = F.interpolate(gt_masks.float().unsqueeze(1), size=(H, W), mode='nearest').squeeze(1).long()
        outputs = self.postOutput(outputs)
        outputs_without_aux = {k: v for k, v in outputs.items() if k != "aux_outputs"}
        targets = self._get_targets(gt_masks)
        # 获取最后一层输出与目标之间的匹配
        indices = self.matcher(outputs_without_aux, targets)

        # 计算所有节点中目标框的平均数量，用于归一化目的
        num_masks = sum(len(t["labels"]) for t in targets)
        num_masks = torch.as_tensor([num_masks], dtype=torch.float, device=next(iter(outputs.values())).device)
        if is_dist_avail_and_initialized():
            torch.distributed.all_reduce(num_masks)
        num_masks = torch.clamp(num_masks / get_world_size(), min=1).item()

        # 计算所有损失
        losses = {}
        for loss in self.losses:
            losses.update(self.get_loss(loss, outputs, targets, indices, num_masks))

        # 如果有辅助损失，则对每个中间层的输出重复此过程。
        if "aux_outputs" in outputs:
            for i, aux_outputs in enumerate(outputs["aux_outputs"]):
                indices = self.matcher(aux_outputs, targets)
                for loss in self.losses:
                    l_dict = self.get_loss(loss, aux_outputs, targets, indices, num_masks)
                    l_dict = {k + f"_{i}": v for k, v in l_dict.items()}
                    losses.update(l_dict)
        loss_ce = 0.0
        loss_dice = 0.0
        loss_mask = 0.0
        for k in list(losses.keys()):
            if k in self.weight_dict:
                losses[k] *= self.weight_dict[k]
                if '_ce' in k:
                    loss_ce += losses[k]
                elif '_dice' in k:
                    loss_dice += losses[k]
                elif '_mask' in k:
                    loss_mask += losses[k]
            else:
                # 如果在 `weight_dict` 中未指定，则删除此损失
                losses.pop(k)
        loss = loss_ce + loss_dice + loss_mask
        losses["loss_tot"] = loss
        return losses

    def _get_binary_mask(self, target):
        y, x = target.size()
        target_onehot = torch.zeros(self.num_classes + 1, y, x).to(target.device)
        target_onehot = target_onehot.scatter(dim=0, index=target.unsqueeze(0), value=1)
        return target_onehot

    def _get_targets(self, gt_masks):
        targets = []
        for mask in gt_masks:
            binary_masks = self._get_binary_mask(mask)
            cls_label = torch.unique(mask)
            labels = cls_label[:] if cls_label[0] != 0 else cls_label[1:]  # TODO for FBP
            binary_masks = binary_masks[labels]
            targets.append({'masks': binary_masks, 'labels': labels})
        return targets

    def __repr__(self):
        head = "Criterion " + self.__class__.__name__
        body = [
            "matcher: {}".format(self.matcher.__repr__(_repr_indent=8)),
            "losses: {}".format(self.losses),
            "weight_dict: {}".format(self.weight_dict),
            "num_classes: {}".format(self.num_classes),
            "eos_coef: {}".format(self.eos_coef),
            "num_points: {}".format(self.num_points),
            "oversample_ratio: {}".format(self.oversample_ratio),
            "importance_sample_ratio: {}".format(self.importance_sample_ratio),
        ]
        _repr_indent = 4
        lines = [head] + [" " * _repr_indent + line for line in body]
        return "\n".join(lines)


class Criterion(object):
    def __init__(self, num_classes, alpha=0.5, gamma=2, weight=None, ignore_index=0):
        self.num_classes = num_classes
        self.alpha = alpha
        self.gamma = gamma
        self.weight = weight
        self.ignore_index = ignore_index
        self.smooth = 1e-5
        self.ce_fn = nn.CrossEntropyLoss(weight=self.weight, ignore_index=self.ignore_index, reduction='none')

    def get_loss(self, outputs, gt_masks):
        """执行损失计算。
        参数:
             outputs: 张量的字典，查看模型的输出规范以了解格式
             gt_masks: [bs, h_net_output, w_net_output]
        """
        loss_labels = 0.0
        loss_masks = 0.0
        loss_dices = 0.0
        num = gt_masks.shape[0]
        pred_logits = [outputs["pred_logits"].float()]  # [bs, num_query, num_classes + 1]
        pred_masks = [outputs['pred_masks'].float()]  # [bs, num_query, h, w]
        targets = self._get_targets(gt_masks, pred_logits[0].shape[1], pred_logits[0].device)
        for aux_output in outputs['aux_outputs']:
            pred_logits.append(aux_output["pred_logits"].float())
            pred_masks.append(aux_output["pred_masks"].float())

        gt_label = targets['labels']  # [bs, num_query]
        gt_mask_list = targets['masks']
        for mask_cls, pred_mask in zip(pred_logits, pred_masks):
            loss_labels += F.cross_entropy(mask_cls.transpose(1, 2), gt_label)
            # loss_masks += self.focal_loss(pred_result, gt_masks.to(pred_result.device))
            loss_dices += self.dice_loss(pred_mask, gt_mask_list)

        return loss_labels / num, loss_dices / num

    def binary_dice_loss(self, inputs, targets):
        inputs = inputs.sigmoid()
        inputs = inputs.flatten(1)
        targets = targets.flatten(1)
        numerator = 2 * torch.einsum("nc,mc->nm", inputs, targets)
        denominator = inputs.sum(-1)[:, None] + targets.sum(-1)[None, :]
        loss = 1 - (numerator + 1) / (denominator + 1)
        return loss.mean()

    def dice_loss(self, predict, targets):
        bs = predict.shape[0]
        total_loss = 0
        for i in range(bs):
            pred_mask = predict[i]
            tgt_mask = targets[i].to(predict.device)
            dice_loss_value = self.binary_dice_loss(pred_mask, tgt_mask)
            total_loss += dice_loss_value
        return total_loss / bs

    def focal_loss(self, preds, labels):
        """
        preds: [bs, num_class + 1, h, w]
        labels: [bs, h, w]
        """
        logpt = -self.ce_fn(preds, labels)
        pt = torch.exp(logpt)
        loss = -((1 - pt) ** self.gamma) * self.alpha * logpt
        return loss.mean()

    def _get_binary_mask(self, target):
        y, x = target.size()
        target_onehot = torch.zeros(self.num_classes + 1, y, x)
        target_onehot = target_onehot.scatter(dim=0, index=target.unsqueeze(0), value=1)
        return target_onehot

    def _get_targets(self, gt_masks, num_query, device):
        binary_masks = []
        gt_labels = []
        for mask in gt_masks:
            mask_onehot = self._get_binary_mask(mask)
            cls_label = torch.unique(mask)
            labels = torch.full((num_query,), 0, dtype=torch.int64, device=gt_masks.device)
            labels[:len(cls_label)] = cls_label
            binary_masks.append(mask_onehot[cls_label])
            gt_labels.append(labels)
        return {"labels": torch.stack(gt_labels).to(device), "masks": binary_masks}



from model.semseg.base import BaseNet

import torch
from torch import nn
import torch.nn.functional as F


class DeepLabV2(BaseNet):
    def __init__(self, backbone, nclass):
        super(DeepLabV2, self).__init__(backbone)

        self.classifier = nn.ModuleList()
        for dilation in [6, 12, 18, 24]:
            self.classifier.append(
                nn.Conv2d(2048, nclass, kernel_size=3, stride=1, padding=dilation, dilation=dilation, bias=True))

        for m in self.classifier:
            m.weight.data.normal_(0, 0.01)

    def base_forward(self, x):
        h, w = x.shape[-2:]

        x = self.backbone.base_forward(x)[-1]

        out = self.classifier[0](x)
        for i in range(len(self.classifier) - 1):
            out += self.classifier[i+1](x)

        out = F.interpolate(out, size=(h, w), mode="bilinear", align_corners=True)

        return out



from collections import OrderedDict
from pprint import pprint

import numpy as np
from prettytable import PrettyTable


def intersect_and_union(pred_label: np.array, label: np.array,
                        num_classes: int, ignore_index: int):
    # Calculate Intersection and Union.

    mask = (label != ignore_index)
    pred_label = pred_label[mask] - 1
    label = label[mask] - 1

    intersect = pred_label[pred_label == label]
    area_intersect, _ = np.histogram(intersect, bins=num_classes, range=(0, num_classes - 1))
    area_pred_label, _ = np.histogram(pred_label, bins=num_classes, range=(0, num_classes - 1))
    area_label, _ = np.histogram(label, bins=num_classes, range=(0, num_classes - 1))

    area_intersect, _ = np.histogram(intersect, bins=num_classes - 1, range=(0, num_classes - 1 - 1))
    area_pred_label, _ = np.histogram(pred_label, bins=num_classes - 1, range=(0, num_classes - 1 - 1))
    area_label, _ = np.histogram(label, bins=num_classes - 1, range=(0, num_classes - 1 - 1))

    area_union = area_pred_label + area_label - area_intersect
    # return area_intersect[1:], area_union[1:], area_pred_label[1:], area_label[1:]
    return area_intersect, area_union, area_pred_label, area_label


def intersect_and_union1(pred_label: np.array, label: np.array,
                         num_classes: int, ignore_index: int):
    # Calculate Intersection and Union.

    mask = (label != ignore_index)
    pred_label = pred_label[mask]
    label = label[mask]

    intersect = pred_label[pred_label == label]
    area_intersect = np.bincount(intersect, minlength=num_classes)
    area_pred_label = np.bincount(pred_label, minlength=num_classes)
    area_label = np.bincount(label, minlength=num_classes)

    area_union = area_pred_label + area_label - area_intersect
    return area_intersect, area_union, area_pred_label, area_label


def total_area_to_metrics(total_area_intersect: np.ndarray,
                          total_area_union: np.ndarray,
                          total_area_pred_label: np.ndarray,
                          total_area_label: np.ndarray,
                          metrics=['mIoU', 'mDice', 'mFscore'],
                          nan_to_num=None,
                          beta: int = 1):
    # Calculate evaluation metrics

    def f_score(precision, recall, beta=1):
        # calculate the f-score value.

        score = (1 + beta ** 2) * (precision * recall) / (
                (beta ** 2 * precision) + recall)
        return score

    if isinstance(metrics, str):
        metrics = [metrics]
    allowed_metrics = ['mIoU', 'mDice', 'mFscore']
    if not set(metrics).issubset(set(allowed_metrics)):
        raise KeyError(f'metrics {metrics} is not supported')

    all_acc = total_area_intersect.sum() / total_area_label.sum()
    ret_metrics = OrderedDict({'aAcc': all_acc})
    for metric in metrics:
        if metric == 'mIoU':
            iou = total_area_intersect / total_area_union
            acc = total_area_intersect / total_area_label
            ret_metrics['IoU'] = iou
            ret_metrics['Acc'] = acc
        elif metric == 'mDice':
            dice = 2 * total_area_intersect / (
                    total_area_pred_label + total_area_label)
            acc = total_area_intersect / total_area_label
            ret_metrics['Dice'] = dice
            ret_metrics['Acc'] = acc
        elif metric == 'mFscore':
            precision = total_area_intersect / total_area_pred_label
            recall = total_area_intersect / total_area_label
            f_value = np.array([
                f_score(x[0], x[1], beta) for x in zip(precision, recall)
            ])
            ret_metrics['Fscore'] = f_value
            ret_metrics['Precision'] = precision
            ret_metrics['Recall'] = recall

    ret_metrics = {
        metric: value
        for metric, value in ret_metrics.items()
    }
    if nan_to_num is not None:
        ret_metrics = OrderedDict({
            metric: np.nan_to_num(metric_value, nan=nan_to_num)
            for metric, metric_value in ret_metrics.items()
        })
    return ret_metrics


def compute_metrics(results: list):
    # Compute the metrics from processed results.

    # convert list of tuples to tuple of lists, e.g.
    # [(A_1, B_1, C_1, D_1), ...,  (A_n, B_n, C_n, D_n)] to
    # ([A_1, ..., A_n], ..., [D_1, ..., D_n])
    results = tuple(zip(*results))
    assert len(results) == 4

    total_area_intersect = sum(results[0])
    total_area_union = sum(results[1])
    total_area_pred_label = sum(results[2])
    total_area_label = sum(results[3])
    ret_metrics = total_area_to_metrics(
        total_area_intersect, total_area_union, total_area_pred_label, total_area_label,
        metrics=['mIoU', 'mDice', 'mFscore'],
        nan_to_num=None,
        beta=1)

    class_names = [str(i) for i in range(24)]
    # summary table
    ret_metrics_summary = OrderedDict({
        ret_metric: np.round(np.nanmean(ret_metric_value) * 100, 5)
        for ret_metric, ret_metric_value in ret_metrics.items()
    })
    metrics = dict()
    for key, val in ret_metrics_summary.items():
        if key == 'aAcc':
            metrics[key] = val
        else:
            metrics['m' + key] = val

    # each class table
    ret_metrics.pop('aAcc', None)
    ret_metrics_class = OrderedDict({
        ret_metric: np.round(ret_metric_value * 100, 2)
        for ret_metric, ret_metric_value in ret_metrics.items()
    })
    ret_metrics_class.update({'Class': class_names})
    ret_metrics_class.move_to_end('Class', last=False)
    class_table_data = PrettyTable()
    for key, val in ret_metrics_class.items():
        class_table_data.add_column(key, val)

    print('per class results:')
    print(class_table_data.get_string())

    return metrics


class Evaluator(object):
    def __init__(self, num_class, ignore_index=-1):
        self.num_class = num_class
        self.ignore_index = ignore_index
        self.batch_results = []

    def reset(self):
        self.batch_results.clear()

    def add_batch(self, gt_image, pre_image):
        # gt [0:n-1] <==> pre [0:n-1]
        assert gt_image.shape == pre_image.shape, f"{gt_image.shape}!={pre_image.shape}"
        self.batch_results.append(intersect_and_union(pred_label=pre_image, label=gt_image,
                                                      num_classes=25, ignore_index=0))


if __name__ == '__main__':
    import os, cv2

    name = "GF2_PMS2__L1A0001119060-MSS2.png"
    name = "GF2_PMS2__L1A0000718813-MSS2.png"
    gt = cv2.imread(os.path.join(r"F:\data_lwc\FBP\test\label", name), 0)
    pr = cv2.imread(os.path.join(r"F:\data_lwc\FBP\test\mask2former\FBP_1024\d2", name), 0)
    arr = []
    for name in os.listdir(r"F:\data_lwc\FBP\test\label"):
        # if name == "GF2_PMS2__L1A0001119060-MSS2.png": continue
        # name = "GF2_PMS2__L1A0001119060-MSS2.png"
        print(name)
        gt = cv2.imread(os.path.join(r"F:\data_lwc\FBP\test\label", name), 0)[54:6800 + 54, 60:7200 + 60]
        pr = cv2.imread(os.path.join(r"D:\git\wbuilding\PFNet_IPNet\result\mask2former\FBP\testcw", name), 0)
        arr.append(intersect_and_union(pred_label=pr, label=gt,
                                       num_classes=25, ignore_index=0))
        # break
    print(len(arr))
    m = compute_metrics(arr)
    pprint(m)



#-*- coding:utf-8 -*-

#############################################
# File Name: setup.py
# Author: mzd
# Mail: 1282032474@qq.com
# Created Time:  2024-10-8
#############################################

from setuptools import setup, find_packages            #这个包没有的可以pip一下

setup(
    name = "segTrain",      #这里是pip项目发布的名称
    version = "1.0.0",  #版本号，数值大的会优先被pip
    keywords = ("pip", "segTrain"),
    description = "llm finetune",

    author = "mzd",
    author_email = "1282032474@qq.com",

    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires = []          #这个项目需要的第三方库
)


# Copyright (c) Facebook, Inc. and its affiliates.
# Modified by Bowen Cheng from https://github.com/facebookresearch/detr/blob/master/models/matcher.py
"""
Modules to compute the matching cost and solve the corresponding LSAP.
"""
import torch
import torch.nn.functional as F
from scipy.optimize import linear_sum_assignment
from torch import nn
from torch.cuda.amp import autocast

from .point_features import point_sample


def batch_dice_loss(inputs: torch.Tensor, targets: torch.Tensor):
    """
    Compute the DICE loss, similar to generalized IOU for masks
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
    """
    inputs = inputs.sigmoid()
    inputs = inputs.flatten(1)
    numerator = 2 * torch.einsum("nc,mc->nm", inputs, targets)
    denominator = inputs.sum(-1)[:, None] + targets.sum(-1)[None, :]
    loss = 1 - (numerator + 1) / (denominator + 1)
    return loss


def batch_sigmoid_ce_loss(inputs: torch.Tensor, targets: torch.Tensor):
    """
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
    Returns:
        Loss tensor
    """
    hw = inputs.shape[1]

    pos = F.binary_cross_entropy_with_logits(
        inputs, torch.ones_like(inputs), reduction="none"
    )
    neg = F.binary_cross_entropy_with_logits(
        inputs, torch.zeros_like(inputs), reduction="none"
    )

    loss = torch.einsum("nc,mc->nm", pos, targets) + torch.einsum("nc,mc->nm", neg, (1 - targets)
                                                                  )

    return loss / hw


def batch_sigmoid_focal_loss(inputs, targets, alpha: float = 0.25, gamma: float = 2):
    """
    Loss used in RetinaNet for dense detection: https://arxiv.org/abs/1708.02002.
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
        alpha: (optional) Weighting factor in range (0,1) to balance
                positive vs negative examples. Default = -1 (no weighting).
        gamma: Exponent of the modulating factor (1 - p_t) to
               balance easy vs hard examples.
    Returns:
        Loss tensor
    """
    hw = inputs.shape[1]

    prob = inputs.sigmoid()
    focal_pos = ((1 - prob) ** gamma) * F.binary_cross_entropy_with_logits(
        inputs, torch.ones_like(inputs), reduction="none"
    )
    focal_neg = (prob ** gamma) * F.binary_cross_entropy_with_logits(
        inputs, torch.zeros_like(inputs), reduction="none"
    )
    if alpha >= 0:
        focal_pos = focal_pos * alpha
        focal_neg = focal_neg * (1 - alpha)

    loss = torch.einsum("nc,mc->nm", focal_pos, targets) + torch.einsum("nc,mc->nm", focal_neg, (1 - targets))

    return loss / hw


class HungarianMatcher(nn.Module):
    """This class computes an assignment between the targets and the predictions of the network
    For efficiency reasons, the targets don't include the no_object. Because of this, in general,
    there are more predictions than targets. In this case, we do a 1-to-1 matching of the best predictions,
    while the others are un-matched (and thus treated as non-objects).
    """

    def __init__(self, cost_class: float = 1, cost_mask: float = 1, cost_dice: float = 1, num_points: int = 0):
        """Creates the matcher

        Params:
            cost_class: This is the relative weight of the classification error in the matching cost
            cost_mask: This is the relative weight of the focal loss of the binary mask in the matching cost
            cost_dice: This is the relative weight of the dice loss of the binary mask in the matching cost
        """
        super().__init__()
        self.cost_class = cost_class
        self.cost_mask = cost_mask
        self.cost_dice = cost_dice

        assert cost_class != 0 or cost_mask != 0 or cost_dice != 0, "all costs cant be 0"

        self.num_points = num_points

    @torch.no_grad()
    def memory_efficient_forward(self, outputs, targets):
        """More memory-friendly matching"""
        bs, num_queries = outputs["pred_logits"].shape[:2]

        indices = []

        # Iterate through batch size
        for b in range(bs):
            out_prob = outputs["pred_logits"][b].softmax(-1)  # [num_queries, num_classes+1]
            out_mask = outputs["pred_masks"][b]  # [num_queries, H_pred, W_pred]

            tgt_ids = targets[b]["labels"]  # [1,2,3, ……]
            tgt_mask = targets[b]["masks"].to(out_mask)  # [c, h, w] c = len(tgt_ids)

            # Compute the classification cost. Contrary to the loss, we don't use the NLL,
            # but approximate it in 1 - proba[target class].
            # The 1 is a constant that doesn't change the matching, it can be ommitted.
            cost_class = -out_prob[:, tgt_ids]  # [num_queries, num_total_targets]

            # ===========================Mask2Former====================================#
            out_mask = out_mask[:, None]  # [num_queries, 1, H_pred, W_pred]
            tgt_mask = tgt_mask[:, None]  # [c, 1, h, w]

            # all masks share the same set of points for efficient matching!
            point_coords = torch.rand(1, self.num_points, 2, device=out_mask.device)
            # get gt labels
            tgt_mask = point_sample(
                tgt_mask,  # [c, 1, h, w]
                point_coords.repeat(tgt_mask.shape[0], 1, 1),  # [c, self.num_points, 2]
                align_corners=False,
            ).squeeze(1)  # [c, self.num_points]

            out_mask = point_sample(
                out_mask,
                point_coords.repeat(out_mask.shape[0], 1, 1),
                align_corners=False,
            ).squeeze(1)  # [num_queries, self.num_points]
            # ===========================end====================================#

            # ===========================MaskFormer====================================#
            # Flatten spatial dimension
            # out_mask = out_mask.flatten(1)  # [num_queries, H*W]
            # tgt_mask = tgt_mask.flatten(1)  # [num_total_targets, H*W]

            with autocast(enabled=False):
                out_mask = out_mask.float()
                tgt_mask = tgt_mask.float()
                # Compute the focal loss between masks
                cost_mask = batch_sigmoid_focal_loss(out_mask, tgt_mask)

                # Compute the dice loss betwen masks
                cost_dice = batch_dice_loss(out_mask, tgt_mask)

            # Final cost matrix
            C = (
                    self.cost_mask * cost_mask
                    + self.cost_class * cost_class
                    + self.cost_dice * cost_dice
            )
            C = C.reshape(num_queries, -1).cpu()  # [num_queries, num_total_targets]

            indices.append(linear_sum_assignment(C))

        return [
            (torch.as_tensor(i, dtype=torch.int64), torch.as_tensor(j, dtype=torch.int64))
            for i, j in indices
        ]

    @torch.no_grad()
    def forward(self, outputs, targets):
        """Performs the matching

        Params:
            outputs: This is a dict that contains at least these entries:
                 "pred_logits": Tensor of dim [batch_size, num_queries, num_classes] with the classification logits
                 "pred_masks": Tensor of dim [batch_size, num_queries, H_pred, W_pred] with the predicted masks

            targets: This is a list of targets (len(targets) = batch_size), where each target is a dict containing:
                 "labels": Tensor of dim [num_target_boxes] (where num_target_boxes is the number of ground-truth
                           objects in the target) containing the class labels
                 "masks": Tensor of dim [num_target_boxes, H_gt, W_gt] containing the target masks

        Returns:
            A list of size batch_size, containing tuples of (index_i, index_j) where:
                - index_i is the indices of the selected predictions (in order)
                - index_j is the indices of the corresponding selected targets (in order)
            For each batch element, it holds:
                len(index_i) = len(index_j) = min(num_queries, num_target_boxes)
        """
        return self.memory_efficient_forward(outputs, targets)

    def __repr__(self, _repr_indent=4):
        head = "Matcher " + self.__class__.__name__
        body = [
            "cost_class: {}".format(self.cost_class),
            "cost_mask: {}".format(self.cost_mask),
            "cost_dice: {}".format(self.cost_dice),
        ]
        lines = [head] + [" " * _repr_indent + line for line in body]
        return "\n".join(lines)



from model.semseg.base import BaseNet

import torch
from torch import nn
import torch.nn.functional as F


class PSPNet(BaseNet):
    def __init__(self, backbone, nclass):
        super(PSPNet, self).__init__(backbone)

        self.head = PSPHead(self.backbone.channels[-1], nclass)


class PSPHead(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(PSPHead, self).__init__()
        inter_channels = in_channels // 4

        self.conv5 = nn.Sequential(PyramidPooling(in_channels),
                                   nn.Conv2d(in_channels * 2, inter_channels, 3, padding=1, bias=False),
                                   nn.BatchNorm2d(inter_channels),
                                   nn.ReLU(True),
                                   nn.Dropout(0.1, False),
                                   nn.Conv2d(inter_channels, out_channels, 1))

    def forward(self, x):
        return self.conv5(x)


class PyramidPooling(nn.Module):
    def __init__(self, in_channels):
        super(PyramidPooling, self).__init__()
        self.pool1 = nn.AdaptiveAvgPool2d(1)
        self.pool2 = nn.AdaptiveAvgPool2d(2)
        self.pool3 = nn.AdaptiveAvgPool2d(3)
        self.pool4 = nn.AdaptiveAvgPool2d(6)

        out_channels = int(in_channels/4)
        self.conv1 = nn.Sequential(nn.Conv2d(in_channels, out_channels, 1, bias=False),
                                   nn.BatchNorm2d(out_channels),
                                   nn.ReLU(True))
        self.conv2 = nn.Sequential(nn.Conv2d(in_channels, out_channels, 1, bias=False),
                                   nn.BatchNorm2d(out_channels),
                                   nn.ReLU(True))
        self.conv3 = nn.Sequential(nn.Conv2d(in_channels, out_channels, 1, bias=False),
                                   nn.BatchNorm2d(out_channels),
                                   nn.ReLU(True))
        self.conv4 = nn.Sequential(nn.Conv2d(in_channels, out_channels, 1, bias=False),
                                   nn.BatchNorm2d(out_channels),
                                   nn.ReLU(True))

    def forward(self, x):
        h, w = x.shape[-2:]
        feat1 = F.interpolate(self.conv1(self.pool1(x)), (h, w), mode="bilinear", align_corners=True)
        feat2 = F.interpolate(self.conv2(self.pool2(x)), (h, w), mode="bilinear", align_corners=True)
        feat3 = F.interpolate(self.conv3(self.pool3(x)), (h, w), mode="bilinear", align_corners=True)
        feat4 = F.interpolate(self.conv4(self.pool4(x)), (h, w), mode="bilinear", align_corners=True)
        return torch.cat((x, feat1, feat2, feat3, feat4), 1)



import os
import numpy as np
from PIL import Image
import cv2
import math
import h5py
import torch
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm

Image.MAX_IMAGE_PIXELS = None


def crop_from_one(image, label, head=(0, 0), size=512):
    x, y = head
    return image[x:x + size, y:y + size, :], label[x:x + size, y:y + size]


def createHDF5(imagePath, labelPath, output_hdf5, downsample_tuple, size=1024):
    print(f"==============[{downsample_tuple}]==============")
    assert len(downsample_tuple) == 3
    d1, d2, d3 = downsample_tuple
    print(imagePath)
    print(labelPath)
    print(output_hdf5)
    ImageNums = len(os.listdir(imagePath))
    oneImage = os.listdir(imagePath)[0]
    H0, W0, _ = cv2.imread(os.path.join(imagePath, oneImage)).shape
    s1 = d1 * size
    s2 = d2 * size
    s3 = d3 * size
    pad_size = (s3 - s1) // 2
    delt23 = (s3 - s2) // 2
    nums = math.ceil(H0 / s1) * math.ceil(W0 / s1) * ImageNums
    with h5py.File(output_hdf5, "w") as hf:
        image1_dataset = hf.create_dataset("image1", shape=(nums, size, size, 3), dtype="uint8")
        image2_dataset = hf.create_dataset("image2", shape=(nums, size, size, 3), dtype="uint8")
        image3_dataset = hf.create_dataset("image3", shape=(nums, size, size, 3), dtype="uint8")

        label_dataset = hf.create_dataset("labels", shape=(nums, size, size), dtype="uint8")
        idx = 0
        for name in tqdm(os.listdir(imagePath)):
            name = name[:-4]
            image0 = cv2.imread(os.path.join(imagePath, name + ".jpg"))
            label0 = np.array(Image.open(os.path.join(labelPath, name + ".png"))).astype(np.uint8)
            image_padded = np.pad(image0,
                                  pad_width=((pad_size, pad_size), (pad_size, pad_size), (0, 0)),
                                  mode='constant')
            # label[label > 0] = 1  # FBP is not need # TODO
            if np.max(label0) > 200:
                label0[label0 > 0] = 1

            for i in range(0, H0, s1):
                i = H0 - s1 if i + s1 > H0 else i
                for j in range(0, W0, s1):
                    j = W0 - s1 if j + s1 > W0 else j
                    img, ann = crop_from_one(image=image0, label=label0, head=(i, j), size=size)
                    image1_dataset[idx] = img
                    image2_dataset[idx] = cv2.resize(
                        image_padded[i + delt23:i + delt23 + s2, j + delt23:j + delt23 + s2, :],
                        (size, size),
                        interpolation=cv2.INTER_LINEAR)
                    image3_dataset[idx] = cv2.resize(image_padded[i:i + s3, j:j + s3, :],
                                                     (size, size),
                                                     interpolation=cv2.INTER_LINEAR)
                    label_dataset[idx] = ann
                    idx += 1
        hf.close()


def TRY():
    os.chdir("try")
    # createHDF5(imagePath=r"F:\data_lwc\FBP\test\img",
    #            labelPath=r"F:\data_lwc\FBP\test\label",
    #            output_hdf5="FBP_d_1_2_4_test.h5",
    #            downsample_tuple=(1, 2, 4),
    #            size=1024)
    hf = h5py.File("FBP_d_1_2_4_test.h5")
    ii = 199
    image1 = hf["image1"][ii]
    image2 = hf["image2"][ii]
    image3 = hf["image3"][ii]

    label = hf["labels"][ii]
    cv2.imwrite("1.jpg", image1)
    cv2.imwrite("2.jpg", image2)
    cv2.imwrite("3.jpg", image3)
    cv2.imwrite("0.png", label)


if __name__ == "__main__":
    # TRY()

    root = r"/scratch/liwanchun"
    path = {"HPD": "HPD",
            "water": "GLH_water",
            "road": "road_data",
            "building": "building_data",
            "FBP": "FBP_data"}
    #
    size = 512  # 1024
    for data_type in ("building", "A", "FBP", "water"):  # ("water", "road", "building"):
        for split in ("train", "val"):
            imagePath = os.path.join(root, path[data_type], split, "img")
            labelPath = os.path.join(root, path[data_type], split, "label")
            assert os.path.exists(imagePath) and os.path.exists(labelPath)
            for d in [(1, 3, 6)]:  # [1, 2, 3, 4, 6, 8]:
                d1, d2, d3 = d
                outPath = os.path.join(root, "Large_Size_Datasets", f"{data_type}_{split}_d_{d1}_{d2}_{d3}_{size}.h5")
                createHDF5(imagePath=imagePath,
                           labelPath=labelPath,
                           output_hdf5=outPath,
                           downsample_tuple=d,
                           size=size)
